; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         HPL.el
; RCS:          $Header: $
; Description:  HPL local settings et alia for HUGE
; Author:       Peter Marvit
; Created:      Fri Apr  6 15:23:46 1990
; Modified:     Fri Mar 22 21:49:41 1991 (Carl Edman) cedman@lynx
; Language:     Emacs-Lisp
; Package:      N/A
; Status:       Experimental (Do Not Distribute)
;
; (C) Copyright 1990, Hewlett-Packard, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Nice macros needed for stuff later on.  Taken from cl.el which is too
;;; big to pre-load.

;;; WHEN and UNLESS
;;; These two forms are simplified ifs, with a single branch.

(defmacro when (condition &rest body)
  "(when CONDITION . BODY) => evaluate BODY if CONDITION is true."
  (list* 'if (list 'not condition) '() body))

(defmacro unless (condition &rest body)
  "(unless CONDITION . BODY) => evaluate BODY if CONDITION is false."
  (list* 'if condition '() body))

;;; The popular c[ad]*r functions.

(defun caar (X)
  "Return the car of the car of X."
  (car (car X)))

(defun cadr (X)
  "Return the car of the cdr of X."
  (car (cdr X)))

(defun cdar (X)
  "Return the cdr of the car of X."
  (cdr (car X)))

(defun cddr (X)
  "Return the cdr of the cdr of X."
  (cdr (cdr X)))

(defun caaar (X)
  "Return the car of the car of the car of X."
  (car (car (car X))))

(defun caadr (X)
  "Return the car of the car of the cdr of X."
  (car (car (cdr X))))

(defun cadar (X)
  "Return the car of the cdr of the car of X."
  (car (cdr (car X))))

(defun cdaar (X)
  "Return the cdr of the car of the car of X."
  (cdr (car (car X))))

(defun caddr (X)
  "Return the car of the cdr of the cdr of X."
  (car (cdr (cdr X))))

(defun cdadr (X)
  "Return the cdr of the car of the cdr of X."
  (cdr (car (cdr X))))

(defun cddar (X)
  "Return the cdr of the cdr of the car of X."
  (cdr (cdr (car X))))

(defun cdddr (X)
  "Return the cdr of the cdr of the cdr of X."
  (cdr (cdr (cdr X))))

(defun caaaar (X)
  "Return the car of the car of the car of the car of X."
  (car (car (car (car X)))))

(defun caaadr (X)
  "Return the car of the car of the car of the cdr of X."
  (car (car (car (cdr X)))))

(defun caadar (X)
  "Return the car of the car of the cdr of the car of X."
  (car (car (cdr (car X)))))

(defun cadaar (X)
  "Return the car of the cdr of the car of the car of X."
  (car (cdr (car (car X)))))

(defun cdaaar (X)
  "Return the cdr of the car of the car of the car of X."
  (cdr (car (car (car X)))))

(defun caaddr (X)
  "Return the car of the car of the cdr of the cdr of X."
  (car (car (cdr (cdr X)))))

(defun cadadr (X)
  "Return the car of the cdr of the car of the cdr of X."
  (car (cdr (car (cdr X)))))

(defun cdaadr (X)
  "Return the cdr of the car of the car of the cdr of X."
  (cdr (car (car (cdr X)))))

(defun caddar (X)
  "Return the car of the cdr of the cdr of the car of X."
  (car (cdr (cdr (car X)))))

(defun cdadar (X)
  "Return the cdr of the car of the cdr of the car of X."
  (cdr (car (cdr (car X)))))

(defun cddaar (X)
  "Return the cdr of the cdr of the car of the car of X."
  (cdr (cdr (car (car X)))))

(defun cadddr (X)
  "Return the car of the cdr of the cdr of the cdr of X."
  (car (cdr (cdr (cdr X)))))

(defun cddadr (X)
  "Return the cdr of the cdr of the car of the cdr of X."
  (cdr (cdr (car (cdr X)))))

(defun cdaddr (X)
  "Return the cdr of the car of the cdr of the cdr of X."
  (cdr (car (cdr (cdr X)))))

(defun cdddar (X)
  "Return the cdr of the cdr of the cdr of the car of X."
  (cdr (cdr (cdr (car X)))))

(defun cddddr (X)
  "Return the cdr of the cdr of the cdr of the cdr of X."
  (cdr (cdr (cdr (cdr X)))))

;;; filename completion to look like Gnuvo
(define-key minibuffer-local-must-match-map " "   'minibuffer-complete)
(define-key minibuffer-local-must-match-map "\^i" 'minibuffer-complete-word)
(define-key minibuffer-local-completion-map " "   'minibuffer-complete)
(define-key minibuffer-local-completion-map "\^i" 'minibuffer-complete-word)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Mandatory Settings
;(setq completion-auto-help nil)		; Don't automatically give help
(setq manual-program "/usr/local/bin/man")	; For Manual-Entry

;;; PATH FOR MH and some necessary settings
(setq mh-progs "/usr/local/bin/mh")	;In case the system we compile on
(setq mh-libs "/usr/local/lib/mh")	;doesn't have MH at that time
(setq mh-redist-full-contents nil)	;Nil since we don't compile with BERK
(setq mh-draft-folder "+drafts")

;; Finally, let's autoload mh-e
;(define-key ctl-x-map "\C-m" 'mh-rmail) ;; starts mhe+inc new mail / +arg scans

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; For HPL lisp items
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *zmacs-compatibility* nil 
"If this is T, then the TMC hacks will try to approximate a lisp
machine.  If it is nil, then it will not try so hard.  Please set this
before loading this file.  (since it is a defvar, it will not override
a previous value)" 
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; AUTOLOADS
;;;**************************************************
;;; Advise (from TMC, ported from CL)
;;;**************************************************
(autoload 'advise "advise" 
"Allows lisp forms to be run before, after, or around the 
execution of another form.  Sort of a super-hook for e-lisp." t)

(autoload 'bibtex-mode "bibtex-mode" "Major mode for editing BibTeX files." t)

;;;**************************************************
;;;  diff
;;;**************************************************
;(autoload 'diff "diff" t "compares any two files or buffers")
;(autoload 'diff-against-reversion "diff" t
;	  "compares a buffer against its backing file")
;(autoload 'diff-previous-version "diff" t
;	  "compares a file against its previous version")


;;;**************************************************
;;; login command
;;;**************************************************
(autoload 'login "login" 
"Start a login session to the given HOST, using a shell-mode buffer.
Possible names come from /etc/hosts.  A single dot means the local host.
Optional flags:  If NOSELECT, don't select the buffer.
If NEW-BUFFER-P (interactive prefix), force creation of a new buffer.
Otherwise, we try to reuse an old buffer, preferring the current buffer."
	   t)

;;;;*********************************************** 
;;;; Dired improvements.
;;;;***********************************************
;
;(autoload 'dired-sort-by-date "dired-resort"
;"In dired, sort the lines by date, newest first.
;With arg, sorts oldest first." t)
;
;(autoload 'dired-sort-by-field "dired-resort"
;"In dired, sort the lines by FIELD (defaults to the mode field)." t)
;
;(autoload 'dired-sort-by-size "dired-resort" 
;"In dired, sort the lines by file size, largest first.  With arg, sorts
;in the reverse order (smallest first).  All directories are grouped
;together at the head of the buffer, and other file types are also
;grouped." t)
;
;(autoload 'dired-sort-by-name "dired-resort"
;"In dired, sort the lines by file name.  With arg, sorts in reverse
;order." t)
;
;(autoload 'dired-sort-by-type "dired-resort" nil t)
;
;(defun setup-dired-mode-map ()
;  (define-key dired-mode-map "D" 'dired-sort-by-date)
;  (define-key dired-mode-map "K" 'dired-sort-by-field)
;  (define-key dired-mode-map "S" 'dired-sort-by-size)
;  (define-key dired-mode-map "N" 'dired-sort-by-name)
;  (define-key dired-mode-map "T" 'dired-sort-by-type)
;  )
;; Must do with advise since dired is autoloaded normally and dired-mode-map
;; is not defined until it is loaded.
;
;(add-load-hook "dired" '(setup-dired-mode-map))
;
(autoload 'complete "completions"
	  "ZMACS-like completions with M-RET as completion cagr.  
           Like dabbrev."  t)

(autoload 'cpp "cpp"
	  "Run cpp on region" t)

;;;; From extras.el   Not all are listed here
(autoload 'backwards-kill-line "extras" nil t)
(autoload 'kill-to-top "extras" nil t)
(autoload 'kill-to-bottom "extras" nil t)
(autoload 'bind-macro "extras" nil t)
(autoload 'lazy-switch-to-buffer "extras" nil t)
(autoload 'slashify-line "extras" nil t)
(autoload 'slashify-region-lines "extras" nil t)

;;;;;End of extras.el

(autoload 'remote-editing "remote" nil t)
(autoload 'find-remote-file "remote" nil t)
(autoload 'save-remote-file "remote" nil t)
(autoload 'write-remote-file "remote" nil t)

(autoload 'repeat "repeat" nil t)

(autoload 'see-chars "see-chars" nil t)

(autoload 'vkill "vkill" 
  "Mode for displaying all UNIX processes owned by the current user (all
  the processes on the system if invoked by the superuser) and allowing
  the user to mark processes to be sent a certain signal.  Processes are
  marked by moving the cursor to the line displaying information about
  the victim process and typing `m' to mark the process.

  If invoked with a prefix arg (optional first arg non-nil from a program)
  the window displaying the process information will be displayed but not
  selected.

  Commands:
    SPC, n  - next line
    DEL, p  - previous line

    m, d    - mark process
    u       - unmark process
    M       - mark all processes
    U       - unmark all processes

    P, g    - update process information
    k, x    - send signal to marked processes (signal read from minibuffer)

    ?       - help"
  t)

(autoload 'list-unix-processes "vkill" 
"List UNIX processes owned by the current user using the ps(1)
command. If run by the superuser, all processes are listed.  The
buffer used to display the listing is put into a special major mode
similar to Dired and Buffer Menu; you can mark processes to be sent
a signal using this buffer.  See the documentation for `vkill-mode'
for more information."
t)

;;;;; From zmacs (not all listed)
(autoload 'comment-out-region "zmacs" 
"Comment out region (with a numeric argument deletes the comment char)
comment-character holds the value of the comment character. (zmacs)"
 t)

(autoload 'shell-command-fast "zmacs" 
  "A quickie way to get shell commands. (zmacs)" 
  t)

(autoload 'create-directory "zmacs" 
  "Create a directory (with mkdir). (zmacs)" 
  t)

(autoload 'delete-directory "zmacs" 
  "Delete an empty directory (with rmdir). (zmacs)" 
  t)

(autoload 'find-unbalanced-parentheses "zmacs" 
  "Check the buffer for unbalanced parentheses.  Stops at any that are 
unbalanced. (zmacs)" 
  t)

(autoload 'split-window-vertically-other-buffer "zmacs" 
  "Split current window into two windows, one above the other.  This
window becomes the uppermost of the two, and gets ARG lines.  No arg
means split equally.  This selects the (other-buffer) in the other
window."
  t)
;;;; END OF zmacs-compatibility

;;; Gnus 3.13
(autoload 'gnus "gnus" "Read network news." t)
(autoload 'gnus-post-news "gnuspost" "Post a new news." t)
(setq gnus-use-generic-from nil)
(setq gnus-default-distribution "") ;;Let inews take care of it.

(autoload 'balance-windows "bal-wins" 
  "Make all visible windows the same size."
  t)

(autoload 'gin-mode "gin-mode" 
  "Gin (Guess Indentation) mode for auto indent" 
  t)

(autoload 'sc-cite-original "supercite" "Hookified Supercite 2.0" t)
(autoload 'sc-cite          "supercite" "Interactive Supercite 2.0" t)
(autoload 'sc-describe      "supercite" "Describe Supercite 2.0" t)

(autoload 'math "math" "Starts Mathematica in a separate buffer" t)
(autoload 'math-mode  "math" 
  "A mode for running Mathematica or editing Mathematica source files." t)
(autoload 'start-math "math" "Starts Mathematica in the current buffer" t)
