; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         HP-softkeys.el
; RCS:          $Header: HP-softkeys.,v 2.0 89/04/01 20:54:32 darrylo Exp $
; Description:  Common code for HP softkeys.  Code is this file is shared
;		between hp-term.el and x-win.el (but this file is loaded
;		from load-terminal-customization in HP-init.el).
; Author:       Darryl Okahata
; Created:      Sat Apr  1 19:37:09 1989
; Modified:     Fri Feb 22 19:41:13 1991
; Language:     Emacs-Lisp
; Package:      N/A
;
; (C) Copyright 1989, 1990, 1991, Hewlett-Packard, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;
;;;	Install new (modified) copy of save-buffers-kill-emacs
;;;

(defun save-buffers-kill-emacs (&optional arg)
  "Offer to save each buffer, then kill this Emacs fork.
With prefix arg, silently save all file-visiting buffers, then kill.
This version turns off HP softkeys before exiting GNU Emacs (loaded from a
GNU Emacs term file)."
  (interactive "P")
  (save-some-buffers arg t)
  (softkeys-off)
  (kill-emacs))


;;#############################################################################

;;Define function key defaults
;;
(if (not (fboundp 'ctl-x-4-tilde-prefix))
  (progn
    (setq ctl-x-4-tilde-map (make-keymap))
    (fset 'ctl-x-4-tilde-prefix ctl-x-4-tilde-map)
    (define-key ctl-x-4-map "~" 'ctl-x-4-tilde-prefix)
  )
)

(defconst default-softkey-1 "\^x4~1")
(defconst default-softkey-2 "\^x4~2")
(defconst default-softkey-3 "\^x4~3")
(defconst default-softkey-4 "\^x4~4")
(defconst default-softkey-5 "\^x4~5")
(defconst default-softkey-6 "\^x4~6")
(defconst default-softkey-7 "\^x4~7")
(defconst default-softkey-8 "\^x4~8")
(defconst default-softkey-9 "\^x4~1")
(defconst default-softkey-10 "\^x4~2")
(defconst default-softkey-11 "\^x4~3")
(defconst default-softkey-12 "\^x4~4")
(defconst default-softkey-13 "\^x4~5")
(defconst default-softkey-14 "\^x4~6")
(defconst default-softkey-15 "\^x4~7")
(defconst default-softkey-16 "\^x4~8")


(defun load-softkey (function-number label function-name &optional
				     string display-labels use-shift)
  "this is an umbrella funtion that will call the appropriate softkey
definition dependant on whether the X11 or terminal interface is being
used.  With this function, a single function key tree definition can
be used for both interfaces.  Note that the X interface has the
capability of shifted softkeys, the terminal interface does not."

  (interactive)
  (if (equal window-system 'x)
      (load-x-function-key function-number label function-name string
			 display-labels use-shift)
    (new-load-function-key function-number label function-name string)))


(defvar x-sfk-off t)


(defun softkeys-on () 
  "Turn on the softkeys label display"
  (interactive)
  (if (equal window-system 'x)
      (progn
	(if x-sfk-off
	    (progn
	      (x-set-sfk-height 2)
	      (setq x-sfk-off nil)))
	(output-x-soft-keys))
    (progn
      (function-keys-on)
      (setq x-sfk-off nil))
  )
  (setq softkeys-arent-displayed nil)
)

(defun set-function-keys-to-nil ()
  ""
  (interactive "")
  (load-softkey 1 " " 'keyboard-quit)
  (load-softkey 2 " " 'keyboard-quit)
  (load-softkey 3 " " 'keyboard-quit)
  (load-softkey 4 " " 'keyboard-quit)
  (load-softkey 5 " " 'keyboard-quit)
  (load-softkey 6 " " 'keyboard-quit)
  (load-softkey 7 " " 'keyboard-quit)
  (load-softkey 8 " " 'keyboard-quit)
)	
	
(defun softkeys-off ()
  "Turn off the softkey label display"
  (interactive)
  (if (equal window-system 'x)
      (progn
        (x-set-sfk-height 0)
        (setq softkeys-arent-displayed t))
    (progn
      (function-keys-off)
      (setq softkeys-arent-displayed t)))
;  (set-function-keys-to-nil) ;added 8/7/89 to define keys to nop
  (setq x-sfk-off t)
)

(defun toggle-softkeys ()
  "Deactivate the softkeys if they're on and turn them on if they're off."
  (interactive)
  (if (equal window-system 'x)
      (x-toggle-softkeys)
    (tm-toggle-softkeys)))
