; Path: hal.com!olivea!uunet!sun-barr!ames!data.nas.nasa.gov!data.nas.nasa.gov!mathur
; From: mathur@nas.nasa.gov (Sanjay R. Mathur)
; Newsgroups: gnu.emacs.sources
; Subject: archie.el with ange-ftp extensions
; Date: 23 Apr 92 06:52:44 GMT
; Organization: NASA Ames Research Center, Moffett Field CA
; 
; 
;      This is an enhanced version of the archie.el posted here some time
; ago. If you have ange-ftp, you can now easily find any of the files named by
; the archie query. 
; 
;      ange-ftp.el is, of course, required but crypt.el is also
; recommended since archive files are usually compressed.
; 
;       For the uninitiated, archie is a program that queries
; archie-servers that maintain databases of files available on anonymous
; ftp sites all over the world. The archie program itself is available
; (It also demonstrates the use of one of the functions of this package;
;    it was inserted by saying M-x archie-sez archie RTN)
; 
; Archie sez: 
; 
; Host ab20.larc.nasa.gov
; 
;     Location: /usenet/comp.sources.misc/volume26
;       DIRECTORY drwxrwxr-x        512  Nov 24 05:01  archie
;     Location: /usenet/comp.sources.misc/volume27
;       DIRECTORY drwxrwxr-x        512  Jan 13 15:50  archie
; 
; Host aix1.segi.ulg.ac.be
; 
;     Location: /pub/docs/tcpip/ftpsites
;            FILE -rw-r--r--      12899  Unk  0 00:00  archie
; 
; Host aix370.rrz.uni-koeln.de
; 
;     Location: /doc/Internet
;       DIRECTORY drwxrwxr-x        384  Jun 22 1991  archie
; 
; Host akiu.gw.tohoku.ac.jp
; 
;     Location: /pub/net
;       DIRECTORY drwxrwxr-x        512  Nov 26 08:44  archie
; 
; Host ashley.cs.widener.edu
; 
;     Location: /pub
;       DIRECTORY drwxr-xr-x        512  Apr  5 13:29  archie
; 
; 
    
;;
;; archie.el v2.0
;;   A mock-interface to Archie for Emacs.
;;
;;   -- original version by Brendan Kehoe (brendan@cs.widener.edu)
;;   ange-ftp extensions by Sanjay Mathur (mathur@nas.nasa.gov)
;; 
;; This file is not part of GNU Emacs but the same permissions apply.
;; 
;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 1, or (at your option)
;; any later version.
;;
;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;;

;; Usage:
;;
;; M-x archie-sez will insert the result of the query in the current buffer.
;; M-x archie creates a separate buffer from which you can find, copy
;; or run dired on any of the entries (using ange-ftp) and redo the search
;; with modified string and/or search-type.
;;

;; Installation instructions:
;;
;; Install this file as archie.el somewhere in your load-path and add the
;; following two lines to ~/.emacs. (without the semicolon's, of course)
;;
;; (autoload 'archie "archie" "Archie interface" t)
;; (autoload 'archie-sez "archie" "Insert results of archie query" t)
;;
;; You may have to change the value of archie-program and archie-server
;; as appropriate for your site.
;; archie-search-type and archie-download-directory can be modified
;; to suit personal preferences.
;;
;; For use with this package, it is also convenient to set
;;  (setq ange-ftp-generate-anonymous-password t)
;;
;;  Also, the crypt package (available in the LCD archives) is useful
;;  with archie-find-file, since most archive sites store their files
;;  in a compressed form.

;;
;; LCD Archive Entry:
;; archie|Sanjay R. Mathur|mathur@nas.nasa.gov|
;; A mock-interface to the archie program.|
;; 92-04-22|2.0|~/interfaces/archie.el.Z|*X|
;;

(require 'ange-ftp)

;; Customization variables

(defvar archie-program "archie"
  "Program that queries archie servers.")

(defvar archie-server "archie.sura.net"
  "Server for \\[archie] searches.

Known archie servers:
	archie.ans.net (USA [NY])
	archie.sura.net (USA [MD])
	archie.mcgill.ca (Canada)
	archie.funet.fi (Finland/Mainland Europe)
	archie.au (Australia)
	archie.doc.ic.ac.uk (Great Britain/Ireland)
")

(defvar archie-download-directory nil
  "Default directory into which any files copied by archie-copy are
  copied. nil means to use the buffer's default-directory.")

(defvar archie-search-type "-e"
  "Search type for \\[archie] searches.
Can be one of: -c for substring searches
               -e for exact matches (default)
               -r for a regexp
               -s for a case-insensitive substring search
")

(defvar archie-l-output "[0-9]*Z *[0-9]* *\\([^ ]*\\) *\\(.*$\\)"
  "Regular expression matching the results of archie -l query. The
   two subexpressions match the host-name and the path respectively.")
;;


(defvar archie-last-query nil)
(defvar archie-last-type nil)

(progn
  (setq archie-mode-map (make-sparse-keymap))
  (define-key archie-mode-map "f" 'archie-find-file)
  (define-key archie-mode-map "a" 'archie)
  (define-key archie-mode-map "a" 'archie-modify-query)
  (define-key archie-mode-map "c" 'archie-copy)
  (define-key archie-mode-map "d" 'archie-dired))

(defun archie-sez (string)
  "Insert the results of an archie query on STRING into the current buffer.
   Uses the function `archie' for its main work."
  (interactive (list (read-string "String: " nil)))
  (archie string nil t))

(defun archie (string &optional type inplace)
  "Look for STRING on an Archie server.
Optional second arg TYPE is the type of search to make -- by default, it's
`archie-search-type'.  Possible values are substring, subcase (case insensitive
substring), and regexp (a regular expression).  Interactively, a prefix arg
will make it prompt for this."
  (interactive (list (read-string "String: " nil)
		     (and current-prefix-arg
			  (read-string "Search type [-c/-e/-r/-s]: "
				       archie-search-type))))
  (or type (setq type archie-search-type))
  (let ((buf (or inplace (get-buffer-create "*Archie*"))))
    (if inplace
	(progn
	  (push-mark (point) t)
	  (insert "Archie sez: \n"))
      (progn
	(setq archie-last-query string)
	(setq archie-last-type type)
	(setq type
	      (if type (concat type "l") "-l"))
	(pop-to-buffer buf)
	(setq buffer-read-only nil)
	(erase-buffer)
	(archie-mode)))
    (call-process archie-program nil buf nil
		  "-h" archie-server
		  type 
		  string))
  (if (not inplace)
      (progn
	(setq buffer-read-only t)
	(goto-char (point-min)))))

(defun archie-get-filename ()
    (beginning-of-line)
    (if (looking-at archie-l-output)
	(concat "/anonymous@"
		(buffer-substring (match-beginning 1) (match-end 1))
		":"
		(buffer-substring (match-beginning 2) (match-end 2)))
      (error "Not archie -l output")))

(defun archie-find-file ()
  "Find the file mentioned on the current line of archie -l output.
   Runs dired if the file is a directory and find-file-run-dired is
   non-nil."
  (interactive)
  (find-file (archie-get-filename)))

(defun archie-copy ()
  "Copy the file mentioned on the current line of archie -l output.
   Prompts with the value of archie-download-directory, if non-nil,
   as the default directory in which to copy. The file-name part can be
   empty, in which case the original name is used."
  (interactive)
  (let* ((from (archie-get-filename))
	 (from-nondir (file-name-nondirectory from))
	 (to nil))
    (if (string-equal "" from-nondir)
	(error "%s is a directory" from))
    (setq to (read-file-name
	      (format "Copy %s to: " from-nondir)
	      archie-download-directory))
    (if (file-directory-p to)
	(setq to (concat (file-name-as-directory to) from-nondir)))
    (copy-file from to 1)))
      


(defun archie-dired ()
  "Run dired on the file or directory mentioned on the current line
   of archie -l output."
  (interactive)
  (dired (file-name-directory (archie-get-filename))))

(defun archie-modify-query (string type)
  "Re-do the last archie search, with modification of the string
   and/or search type."
  (interactive (list (read-string "String: " archie-last-query)
			  (read-string "Search type [-c/-e/-r/-s]: "
				       archie-last-type)))
  (archie string type))

(defun archie-mode ()
  "Major mode for interacting with the archie program.
   Type: \\[archie-find-file]  to find the file on the current line,
   or:  \\[archie-copy] to copy it
   or:  \\[archie-dired] to run dired.

   To redo the last search with modification of the string and/or
   switches, type: \\[archie-modify-query].

   If archie-download-directory is set to non-nil then its value is used
   as the default directory while prompting for the target file by the
   archie-copy command.
   "
  (kill-all-local-variables)
  (rename-buffer (format "Archie [%s] %s"
			    archie-last-type archie-last-query))
  (setq mode-name "Archie")
  (setq major-mode 'archie-mode)
  (use-local-map archie-mode-map))
