;;
;; auto-indent.el -- Set up minor mode to auto-indent relatively.
;; Alan K. Stebbens  7 Aug 87
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; From aks@hub.ucsb.edu  Fri Aug  7 15:07:23 1987
;; Received: by PREP.AI.MIT.EDU; Fri, 7 Aug 87 15:07:23 EDT
;; Message-Id: <8708071907.AA07468@prep.ai.mit.edu>
;; Received: from  by hub.ucsb.edu.ucsb (5.51) id AA16783; Fri, 7 Aug 87 10:35:03 PDT
;; Posted-Date: Fri, 07 Aug 87 10:35:01 PDT
;; To: info-gnu-emacs@prep.ai.mit.edu
;; Subject: auto-indent minor mode
;; Date: Fri, 07 Aug 87 10:35:01 PDT
;; From: Alan Stebbens (aks@hub.ucsb.edu) <aks@hub.ucsb.edu>
;; 
;; 
;; It always seemed to me that auto-indention should be a minor mode,
;; instead of a major mode, as it is with indented-text mode.  So, I
;; wrote the following little piece of Emacs-lisp to provide
;; auto-indention as a minor mode to the current major mode.  This is
;; particularly useful with outline-mode.  It sets the TAB key to
;; indent-relative, and links indent-relative-maybe to the
;; indent-line-function. 
;; 
;; To use: (auto-indent-mode ARG) will set, clear, or toggle the
;; auto-indent-flag if ARG is nonzero, zero, or omitted (null).  As a
;; matter of course, I added the following line to my text-mode-init func
;; (which is invoked via text-mode-hook):
;; 
;;        (auto-indent-mode 1)
;; 
;; If there are improvements which can be made on this, please mail
;; them to me so I can learn from it.
;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; The function 'make-local-vars was added by Darryl Okahata,
;; as it was originally missing from this file.
(defun make-local-vars (&rest list)
  "Make every variable listed in LIST a variable local to a buffer.
LIST is of the form:
        (VAR1 VALUE1 VAR2 VALUE2 ...)
Each VARn is the name of the variable to make into a local variable, and
the corresponding VALUEn is the initial value to give to the local
variable.
"
  (interactive "")
  (let (var value)
    (while list
      (setq var (car list))
      (setq value (car (cdr list)))
      (make-local-variable var)
      (set var value)
      (setq list (cdr (cdr list)))
    )
  )
)

;;
(if (boundp 'auto-indent-flag)
    nil
  (defvar auto-indent-flag nil "\
If set, indicates that auto-indent mode is active.  This variable is 
automatically set by invoking \\[auto-indent-mode].")
  (setq minor-mode-alist (cons '(auto-indent-flag " Indent") 
			       minor-mode-alist))
  (make-variable-buffer-local 'auto-indent-flag)
  (set-default 'auto-indent-flag nil))

;; auto-indent-mode 

(defun auto-indent-mode (arg) "\
Enable, disable, or toggle auto-indent mode if ARG is nonzero, zero, or
omitted.  Auto-indent mode works by invoking indent-relative for TAB,
and using indent-relative-may as the indent-line-function for auto-fill,
and NEWLINE."  
  (interactive "P")
  (if (setq auto-indent-flag (if (or (zerop (prefix-numeric-value arg))
				     (and (null arg) auto-indent-flag))
				 nil 
			       t))
      (progn
	(if (not (boundp 'auto-indent-old-tab-key))
	    (make-local-vars 'auto-indent-old-tab-key (key-binding "\t")
			     'auto-indent-old-indent-func indent-line-function
			     'auto-indent-flag t))
	(local-set-key "\t" 'indent-relative)
	(setq indent-line-function 'indent-relative-maybe))
    (if (boundp 'auto-indent-old-tab-key)
	(progn
	  (local-set-key "\t" auto-indent-old-tab-key)
	  (setq indent-line-function auto-indent-old-indent-func)
	  (makunbound auto-indent-old-tab-key)
	  (makunbound auto-indent-old-indent-func))))
  (set-buffer-modified-p (buffer-modified-p)))


