;From arpa-unix-emacs-request@ALEXANDER.BBN.COM Mon Jan 25 23:59:19 1988
;Received: from alexander by ALEXANDER.BBN.COM id aa03336; 25 Jan 88 22:42 EST
;Received: from [128.89.0.122] by ALEXANDER.BBN.COM id aa03332;
;          25 Jan 88 22:42 EST
;Received: from ucbvax.berkeley.edu by BBN.COM id aa20331; 25 Jan 88 22:40 EST
;Received: by ucbvax.Berkeley.EDU (5.58/1.26)
;	id AA19317; Mon, 25 Jan 88 16:06:14 PST
;Received: from USENET by ucbvax.Berkeley.EDU with netnews
;	for arpa-unix-emacs@bbn.com (unix-emacs@bbn.com)
;	(contact usenet@ucbvax.Berkeley.EDU if you have questions)
;Date: 25 Jan 88 22:39:38 GMT
;From: "Vincent P. Broman" <broman@COD.NOSC.MIL>
;Organization: Naval Ocean Systems Center, San Diego
;Subject: Elisp manual
;Message-Id: <962@cod.NOSC.MIL>
;Sender: unix-emacs-request@BBN.COM
;To: unix-emacs@BBN.COM
;
;Every once in a while, one sees yet another plea for general documentation
;on GNU Emacs Elisp programming.  LaLiberte, Krawitz, and Marick have
;posted very useful articles about it in the past, but nothing supplied
;capsule documentation on the panoply of elisp functions available to
;the programmer.  The quick hack appended to this article produces a buffer
;full of such documentation when pointed at a directory full of emacs
;C source or elisp source.  It just snarfs documentation strings, but I found
;it useful.
;
;No flames regarding the gross-hack-ness of the code, please.
;Also, I suspect it conses something aweful, so save results in a file,
;don't rerun it.
;
;Vincent Broman,  code 632, Naval Ocean Systems Center, San Diego, CA 92152, USA
;Phone: +1 619 553 1641    Internet: broman@nosc.mil   Uucp: sdcsvax!nosc!broman
;
;======================================================================
;Author: Vincent Broman, broman@nosc.mil, Dec 1987.
;Code in the public domain.  That means "no copyright" you non-cognoscenti.

(defun c-src-doc (dirname)
  "Searches given directory for C source files and collects the doc
strings from the functions defined in each .c file,
formatting it into a pretty report.
Doc string must be available online, not #ifdefed out."
  (interactive "Dname of C source directory: ")
  (let ((flis (directory-files dirname nil ".*\\.c$"))
	(docbuf (get-buffer-create "*make doc text*"))
	(srcbuf (get-buffer-create "*get doc sources*"))
	(oldbuf (current-buffer))
	(doc-last-fn-got "not-a-function"))
    (set-buffer srcbuf)
    (erase-buffer)
    (c-mode)
    (set-buffer docbuf)
    (erase-buffer)
    (insert "Emacs-Lisp defuns written in C in directory " dirname "\n")
    (insert "\t\t\tas available in\n ")
    (insert (emacs-version) "\n\n\n\n")
    (while flis
      (insert (doc-1-c-file dirname (car flis) srcbuf))
      (setq flis (cdr flis)))
    (goto-char (point-min))
    (replace-string " nargs args)" " args...)   ")
    (set-buffer oldbuf)
    (kill-buffer srcbuf)
    (message (concat "c-src-doc finished in buffer "
		     (buffer-name docbuf)
		     "."))))

(defun doc-1-c-file (dirname filename srcbuf)
  (let ((docstr "")
	(oldbuf (current-buffer))
	(fullname (concat dirname filename)))
    (if (file-readable-p fullname)
	(progn
	  (message (concat "searching " fullname " ..."))
	  (set-buffer srcbuf)
	  (erase-buffer)
	  (insert-file-contents fullname)
	  (goto-char (point-max))
	  (while (re-search-backward "^DEFUN[ \t]*([ \t]*\"" (point-min) t)
	    (goto-char (match-end 0))
	    (setq docstr (concat
			   (doc-name-to-text
			     filename
			     (buffer-substring (point)
					       (save-excursion
						 (skip-chars-forward "^\"")
						 (point)))
			     (doc-c-style-arglist))
			   docstr))
	    (beginning-of-line))
	  (set-buffer oldbuf)))
    docstr))

(defun doc-c-style-arglist ()		;point sits at start of funname str
  (save-excursion
    (search-backward "(")
    (forward-sexp 1)
    (if (looking-at "[ \t\f\n]*\\*/")
	""
      (progn
	(forward-sexp 1)
	(doc-no-commas (1+ (save-excursion (backward-sexp 1) (point)))
		       (1- (point)))))))

(defun doc-no-commas (startpt endpt)
  (subst-char-in-region startpt endpt ?\, ?\ )
  (subst-char-in-region startpt endpt ?\_ ?\-)
  (goto-char startpt)
  (if (/= ?\) (char-after (point)))
      (progn
	(forward-sexp 1)
	(while (/= ?\) (char-after (point)))
	  (just-one-space)
	  (forward-sexp 1))))
  (buffer-substring startpt (point)))


(defun lisp-src-doc (dirname)
  "Searches given directory for Elisp source files and collects the doc
strings from the functions defined in each .el file,
formatting it into a pretty report."
  (interactive "Dname of Elisp source directory: ")
  (let ((flis (directory-files dirname nil ".*\\.el$"))
	(docbuf (get-buffer-create "*make doc text*"))
	(srcbuf (get-buffer-create "*get doc sources*"))
	(oldbuf (current-buffer))
	(doc-last-fn-got "not-a-function"))
    (set-buffer srcbuf)
    (erase-buffer)
    (emacs-lisp-mode)
    (set-buffer docbuf)
    (erase-buffer)
    (insert "Emacs-Lisp defuns written in Lisp in directory " dirname "\n")
    (insert "\t\t\tas available in\n ")
    (insert (emacs-version) "\n\n\n\n")
    (while flis
      (insert (doc-1-lisp-file dirname (car flis) srcbuf))
      (setq flis (cdr flis)))
    (set-buffer oldbuf)
    (kill-buffer srcbuf)
    (message (concat "lisp-src-doc finished in buffer "
		     (buffer-name docbuf)
		     "."))))

(defun doc-1-lisp-file (dirname filename srcbuf)
  (let ((docstr "")
	(oldbuf (current-buffer))
	(fullname (concat dirname filename))
	defun-fnname
	defun-arglist)
    (if (file-readable-p fullname)
	(progn
	  (message (concat "searching " fullname " ..."))
	  (set-buffer srcbuf)
	  (erase-buffer)
	  (insert-file-contents fullname)
	  (goto-char (point-min))
	  (while (re-search-forward "^(defun " (point-max) t)
	    (goto-char (match-end 0))
	    (setq defun-fnname (read (current-buffer)))
	    (setq defun-arglist (read (current-buffer)))
	    (setq docstr (concat
			   docstr
			   (doc-name-to-text
			     filename
			     (prin1-to-string defun-fnname)
			     (doc-lisp-style-arglist defun-arglist))))
	    (end-of-line))
	  (set-buffer oldbuf)))
    docstr))

(defun doc-lisp-style-arglist (arglist)
  (if arglist
      (if (listp arglist)
	  (let ((str ""))
	    (while arglist
	      (setq str (concat str " " (prin1-to-string (car arglist))))
	      (setq arglist (cdr arglist)))
	    (substring str 1))
	"args...")
    ""))

(defun doc-name-to-text (filename funname arglist)
  (let ((funsymbol (intern-soft funname)))
    (if (and
	  funsymbol
	  (fboundp funsymbol)
	  (not (string-equal funname doc-last-fn-got)))
	(let ((fundoc (documentation funsymbol)))
	  (setq doc-last-fn-got funname)
	  (if (and fundoc (/= 0 (length fundoc)))
	      (doc-format filename funname arglist fundoc)
	    (let (doc-nxt-str)
	      (if (and
		    (eq major-mode 'emacs-lisp-mode)
		    (setq doc-nxt-str (read (current-buffer)))
		    (stringp doc-nxt-str)
		    (/= 0 (length doc-nxt-str)))
		  doc-nxt-str
		""))
	    ""))
      "")))

(defun doc-format (filename funname arglist fundoc)
  (concat "======  ("
	  (setq funname (concat funname
				(if (string-equal arglist "") "" " ")))
	  arglist ")"
	  (make-string (max 4 (- 40 (length funname) (length arglist)))
		       ?\ )
	  filename
	  "\n"
	  fundoc
	  (if (= ?\n (elt (substring fundoc -1 nil) 0))
	      "\n\n"
	    "\n\n\n")))

