;;; -*- Emacs-Lisp -*- 
;;; crypt++-insert.el -- an insert-file that decodes and decrypts
;;; 
;;; $Modified: Sat Feb  6 16:22:58 1993 by dodd $
;;; $Id: crypt++-insert.el,v 1.14 1993/02/06 21:23:01 dodd Exp $ 
;;; $Revision: 1.14 $

;;; Copyright (C) 1993 Rod Whitby and Lawrence R. Dodd
;;; see copyright notice in crypt++.el 

;;; Uses crypt++.el so that insert-file (C-x i) can decode and decrypt.

;;; This program is still being tested and will bundled with crypt++.el in the
;;; future. please send bugs to dodd@roebling.poly.edu.
 
;;; INSTALLATION 
;;; 
;;; To use this package, simply put it in a file called "crypt++-insert.el" in 
;;; a directory known to GNU Emacs, byte-compile it, and put the line
;;;
;;;                      (require 'crypt++-insert)
;;;
;;; in your ~/.emacs file or in the file default.el in the ../lisp directory 
;;; of the GNU Emacs distribution.  This will rebind `C-x i.'

(require 'crypt++)

(defvar crypt-auto-decode-insert nil
  "*t value means that the files inserted into buffers using crypt-insert-file
will be decoded automatically, without requesting confirmation from the user.
`nil' means to ask before doing the decoding.")

;;; attempt at a crypt++ replacement for insert-file-contents
(defun crypt-insert-file-contents (file &optional visit)
  "Acts much like the insert-file-contents except it also decodes
and or decrypts the contents of file prior to inserting, if necessary, using
crypt++.el.  Does not completely mimic insert-file-contents.  For instance, if
visit is t then care must be taken to set the buffer-file-name and make the
file unmodified which is not done below."
  
  (let (temp-buffer
        temp-list
        (crypt-auto-decode-buffer crypt-auto-decode-insert)
        (orig-buffer (current-buffer)))
    
    ;; create a temporary buffer and decode and decrypt it
    (save-excursion
      
      ;; temporary buffer (use the same name as the file to be inserted)
      (setq temp-buffer (generate-new-buffer (file-name-nondirectory file)))
      (set-buffer temp-buffer)
      
      ;; original insert-file-contents - save list 
      (setq temp-list (insert-file-contents file visit))

      ;; make temp-buffer unmodified
      (set-buffer-modified-p nil)
      
      ;; need to set buffer name to file name for crypt++
      (setq buffer-file-name file)
      
      ;; decode and decrypt (if necessary)
      (crypt-find-file-hook)
      
      ;; find the length of the file to be inserted. insert-file-contents 
      ;; returns it for the original encoded/encrypted file 
      (setcdr temp-list (cons (buffer-size) ()))
      
      ;; now insert temp-buffer into original buffer
      (set-buffer orig-buffer)
      (insert-buffer temp-buffer)
      
      ;; kill the temporary buffer
      (kill-buffer temp-buffer))
    
    ;; return (modified) list from original-insert-file-contents
    temp-list))

;;; crypt replacement for insert-file
(defun crypt-insert-file (filename)
  "Insert contents of file FILENAME into buffer after point.
Set mark after the inserted text.  Decrypts and decodes if necessary.  Calls
crypt-insert-file-contents."
  (interactive
   (list
    (read-file-name "Insert File (via crypt++): " default-directory nil t)))
  (let ((tem (crypt-insert-file-contents filename)))
    (push-mark (+ (point) (car (cdr tem))))))

;;; bind crypt-insert-file over the usual insert-file 
(define-key ctl-x-map "i" 'crypt-insert-file)

;;; provide package 
(provide 'crypt++-insert)
