;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         header3.el
; RCS:		$Header$
; Description:  NMODE style standard headers. 
; Author:       Jerry Duggan, HP/FSD  
; Created:      Thu May 29 11:27:45 1986 
; Modified:     Sat Jul  2 01:19:35 1994 (Christian Limpach) chris@nice.ch
; Language:     Emacs-Lisp
; Package:      
; Status:       Experimental (Do Not Distribute)
;
; (c) Copyright 1986, 1987, 1988, 1989, 1990, 1991, Hewlett-Packard, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; buffer specific variables

(make-variable-buffer-local 'default-header-comment-character)
(make-variable-buffer-local 'header-comment-character)
(set-default 'default-header-comment-character ?\#)

(make-variable-buffer-local 'default-header-comment-marker)
(make-variable-buffer-local 'default-header-comment)
(make-variable-buffer-local 'header-comment-marker)
(make-variable-buffer-local 'header-comment)
(make-variable-buffer-local 'header-prefix)
(make-variable-buffer-local 'header-suffix)

(set-default 'default-header-comment-marker "#")
(set-default 'default-header-comment "#")
(set-default 'header-prefix nil)
(set-default 'header-suffix nil)

;; Nice defvars for various globals

(defvar header-line-search 3
  "*Number of lines to search for standard headers")
(defvar header-max-size 50
  "*Maximum size of a header")
(defvar header-author-name nil
  "*Default author name for standard headers")
(defvar header-user-name nil
  "*Default user name for standard headers")
(defvar header-user-company-name nil
  "*Default users company name for standard headers")
(defvar header-copyright-notice nil
  "*Used to override automatic construction of copyright notice in standard headers.")
(defvar header-source-control-name "RCS"
  "Name to put in source control line of standard header; i.e. SCCS or RCS")
(defvar header-source-control-string (concat "$" "Header: " "$") ;dont let rcs get at this string
  "default revision string for source control; i.e. $Header: header.el,v 2.0 89/03/30 12:26:24 darrylo Exp $ or %A% %G% %U%")
;; c@nice.ch added mail address stuff
(defvar header-mail-address nil
  "*Default mail address for standard headers")

(defvar buffer-header-defaults
  '((lisp-mode . (";" "; -*-Lisp-*-" nil ";"))
    (lisp-interaction-mode . (";" "; -*-Emacs-Lisp-*-" nil ";"))
    (emacs-lisp-mode . (";" "; -*-Emacs-Lisp-*-" nil ";"))
    (scheme-mode . (";" "; -*-Scheme-*-" nil ";"))
    (c-mode . (" *" "/* -*-C-*-" " */" " *"))
    (plain-TeX-mode . ("%" "% -*-TeX-*-" nil "%"))
    (plain-tex-mode . ("%" "% -*-TeX-*-" nil "%"))
    (TeX-default-mode . ("%" "% -*-TeX-*-" nil "%"))
    (LaTeX-mode . ("%" "% -*-LaTeX-*-" nil "%"))
    (latex-mode . ("%" "% -*-LaTeX-*-" nil "%"))
    (c++-mode . (" *" "/* -*-C++-*-" " */" " *"))
    (objc-mode . (" *" "/* -*-ObjC-*-" " */" " *"))
    (perl-mode . ("#" "#!/usr/local/bin/perl" nil "#"))
    (awk-mode . ("#" "#!/usr/local/bin/gawk -f" nil "#"))
    (nroff-mode . (".\\\"" ".\\\" -*-Nroff-*-" nil ".\\\" -"))
    )
  "*Alist of default variables for header stuff.  Each element is of the form:
  (<mode> . (<lineprefix> <prefix> <suffix> <linechar>))
    <mode> is the mode of the buffer (lisp-mode, c-mode, etc).
    <lineprefix> is the line-prefix.
    <prefix> is the header-prefix.
    <suffix> is the header-suffix.
    <linechar> is the character appended to <lineprefix> to make a
               separator line.")

(defun header-marker-string (marker)
  (concat
   marker
   (make-string
    (- 79 (length marker))
    (string-to-char
     (substring marker
		(- (length marker) 1)
		(length marker))))))

;;; General predicates for lines and headers

; Returns T if LINE is a marker line
(defun header-marker-line-p (line)
  (save-excursion
    (if (eq (goto-line line) 0)
	(if (let ((p-c (preceding-char)))
	      (or (= p-c 0)
		  (= p-c ?\n)))
	    (let* ((pt (point))
		   (pt80 (+ pt 79))
		   (tmp-h-c-m (or
			       header-comment-marker
			       (let ((p-c (progn (end-of-line)
						 (preceding-char))))
				 (backward-char)
				 (while (and (< pt (point))
					     (eq p-c (preceding-char)))
				   (setq p-c (preceding-char))
				   (backward-char))
				 (if (< pt (point))
				     (buffer-substring pt (+ 1 (point)))
				   (char-to-string (char-after (point))))
				 )
			       )))
	      (if (< pt80 (point-max))
		  (let* ((str (buffer-substring pt pt80))
			 (cmp-str
			  (header-marker-string tmp-h-c-m)
			  ))
		    (and
		     (string-equal str cmp-str)
		     (setq header-comment-marker tmp-h-c-m)))
		nil))
	  nil)
      'error)
    ))

; Returns T if line is a blank comment line

(defun header-blank-comment-line-p (line)
  (save-excursion
    (if (eq (goto-line line) 0)
	(and (let ((p-c (preceding-char)))
	       (or (= p-c 0)
		   (= p-c ?\n)))
	     (string-equal (buffer-substring
			    (point) (+ (point)
				       (length header-comment)))
			   header-comment)
	     )
      'error)))

; Is true if the buffer has a standard header

(defun buffer-has-header-p ()
  (save-excursion
    (let ((header-start (find-first-header-line)))
      (if header-start
	  (let* ((string-begin
		  (progn
		    (goto-line (+ header-start 2))
		    (point)))
		 (string-end (+ string-begin 79)))
	    (string-match "File:"
			  (buffer-substring string-begin string-end)))
	(find-last-header-line)
	(progn
	  (setq header-comment-marker default-header-comment-marker)
	  (setq header-comment default-header-comment)
	  nil)))))

;;; Locators for first, last header line

; Returns the line number of the first header line

(defun find-first-header-line ()
  (catch 'header
    (let ((i 1))
      (while (< i header-line-search)
	(let ((h-m-l (header-marker-line-p i)))
	  (cond ((eq h-m-l 'error)
		 (throw 'header nil))
		((and h-m-l
		      (save-excursion
			(goto-line (1+ i))
			(if header-comment
			    (string-equal
			     header-comment
			     (buffer-substring
			      (point)
			      (+ (point) (length header-comment))))
			  (let ((h-c (buffer-substring
				      (point)
				      (point-at-end-of-line))))
			    (forward-line)
			    (and (string-equal
				  h-c
				  (buffer-substring
				   (point)
				   (+ (point) (length h-c))))
				 (setq header-comment h-c))))))
		 (throw 'header i))
		(t (setq i (1+ i))))))
      nil))
  )

; Returns the line number of the last header line

(defun find-last-header-line ()
  (catch 'header
    (let ((f-h-l (find-first-header-line)))
      (if f-h-l
	(let ((i (1+ f-h-l))
	      (max-search-line (+ f-h-l header-max-size)))
	  (while (< i max-search-line)
	    (let ((h-m-l (header-marker-line-p i)))
	      (cond ((eq h-m-l 'error)
		     (throw 'header nil))
		    (h-m-l
		     (throw 'header i))
		    (t (setq i (1+ i))))))
	  nil)
	nil))))

;;; utility functions

(defun line-to-char-offset (line)
  (goto-line line)
  (point))

;;; Make-Header cookies

(defun insert-attribute-line (prefix attr spaces value)
  (insert-string (format "%s %s: %s%s\n" prefix attr spaces value)))

(defun make-header ()
  "Makes a header in the current buffer, if there is none already there"
  (interactive)
  (set-buffer-header-vars)
  (if (buffer-has-header-p)
      (error "file already has standard header")
    (let ((header-string (concat
			  (header-marker-string header-comment-marker)
			  "\n"))
	  (blank-string (concat header-comment "\n"))
	  old-point)
      ;;      (aset blank-string 0 header-comment-character)
      ;;      (aset blank-string 1 ?\n)
      (goto-line 1)
      (setq old-point (point))
      (if header-prefix
	  (progn
	    (insert-string header-prefix)
	    (insert-string "\n")))
      (insert-string header-string)
      (insert-string blank-string)
      (insert-attribute-line header-comment "File" "        "
			     (file-name-nondirectory (or
						      (buffer-file-name)
						      "")))
      (insert-attribute-line header-comment 
			     header-source-control-name "         "
			     header-source-control-string)
      (insert-attribute-line header-comment "Description" " " "")
      (insert-attribute-line header-comment "Author" "      "
			     (concat
			      (or header-author-name
				  header-user-name
				  "")
			      (if header-mail-address
				  (concat " <" header-mail-address ">"))
			      ))
      (insert-attribute-line header-comment "Created" "     "
			     (current-time-string))
      (insert-attribute-line header-comment "Modified" "    " "")
      (insert-attribute-line header-comment "Language" "    "
			     (if (string= mode-name "Fundamental")
				 "N/A"
			       mode-name))
      (insert-attribute-line header-comment "Package" "     " "N/A")
      (insert-attribute-line header-comment "Status" "      "
			     "Experimental (Do Not Distribute)")
      (insert-string blank-string)
      (insert-string (format "%s (C) Copyright %s, %s%s\n"
			     header-comment
			     (substring (current-time-string) -4)
			     (or header-copyright-notice
				 header-user-company-name				 header-user-name
				 " ")
			     (if header-copyright-notice
				 ""
			       ", all rights reserved.")))
      (insert-string blank-string)
      (insert-string header-string)
      (if header-suffix
	(progn
	  (insert-string header-suffix)
	  (insert-string "\n")))
      (insert-string "\n")
      (goto-char (string-match "Description"
			       (buffer-substring old-point (point))))
      (end-of-line)
      )))

; Update the modify time of a standard header

(defun header-update-modify-time ()
  (set-buffer-header-vars)
  (unwind-protect
      (if (buffer-has-header-p)
	  (save-excursion
	    (let ((header-begin nil)
		  (header-end nil)
		  (mod-field nil)
		  (modify-position nil))
	      (setq header-begin (line-to-char-offset
				  (find-first-header-line)))
	      (setq header-end   (line-to-char-offset
				  (find-last-header-line)))
	      (if (and header-begin header-end)
		  (progn
		    (setq mod-field
			  (string-match
			   (format "%s Modified:" header-comment)
			   (buffer-substring header-begin
					     header-end)))
		    (if (and mod-field
			     (< header-begin mod-field)
			     (< mod-field header-end))
			(progn
			  (setq modify-position
				(+ mod-field (1- header-begin)))
			  (if modify-position
			      (progn
				(goto-char (- modify-position 10))
				(if (re-search-forward " Modified:.*\n"
						       nil t)
				    (replace-match 
				     (format
				      " Modified:     %s\n"
				      (if header-user-name
					  (format "%s (%s) %s"
						  (current-time-string)
						  header-user-name
						  (if header-mail-address
						      header-mail-address
						    (format "%s@%s"
							    (user-login-name)
							    (system-name))))
					(current-time-string))) t t)))
			    nil))
		      nil))
		nil))))))
    
(setq write-file-hooks (cons 'header-update-modify-time write-file-hooks))

;;; Make-Revision cookies

; Will position to a newly made revision entry, making the history if necessary

(defun header-find-or-make-revision-history ()
  (set-buffer-header-vars)
  (if (buffer-has-header-p)
      (let ((last-line (find-last-header-line))
	    (revision-string (format "%s Revisions:\n"
				     header-comment)))
	(if (and (header-blank-comment-line-p (1+ last-line))
		 (header-blank-comment-line-p (+ last-line 3))
		 (string-equal revision-string
			       (buffer-substring
				(line-to-char-offset (+ last-line 2))
				(line-to-char-offset (+ last-line 3)))))
	    (goto-line (+ last-line 4))
	  (let ((header-string (make-string 80 header-comment-character))
		(blank-string (make-string 2 ?\ )))
	    (aset header-string 79 ?\n)
	    (aset blank-string 0 header-comment-character)
	    (aset blank-string 1 ?\n)
	    (goto-line (+ last-line 1))
	    (insert-string blank-string)
	    (insert-string revision-string)
	    (insert-string blank-string)
	    (insert-string header-string)
	    (goto-line (+ last-line 4)))))
    (error "buffer does not have a standard header")))

(defun make-revision ()
  "Make a revision entry, creating the history entry if needed\
Buffer must have a standard header"
  (interactive)
  (set-buffer-header-vars)
  (header-find-or-make-revision-history)
  (insert-string (format "%s %s (%s) %s@%s\n%c  \n"
			 header-comment
			 (current-time-string)
			 (or header-user-name
			     "")
			 (user-login-name)
			 (system-name)
			 header-comment-character))
  (if auto-fill-hook
    (progn
      (setq fill-prefix (format "%s  " header-comment))
      (message "fill prefix is now '%s'" fill-prefix))
    nil)
  (backward-char))

; language customizations:

(defun set-buffer-header-vars ()
  (let ((mode-vars
	 (cdr (assoc (setq mm major-mode) buffer-header-defaults))))
    (if mode-vars
	(setq default-header-comment (car mode-vars)
	      header-prefix (car (cdr mode-vars))
	      header-suffix (car (cdr (cdr mode-vars)))
	      default-header-comment-marker (car (cdr (cdr (cdr mode-vars)))))
      (setq header-comment-marker default-header-comment-marker
	    header-comment default-header-comment))))

(setq find-file-hooks
  (cons 'set-buffer-header-vars find-file-hooks))

(defun get-header-info ()
  "Return a list of attribute-value pairs from the current buffer's file
header, if it has a header."
  (interactive)
  (set-buffer-header-vars)
  (if (buffer-has-header-p)
      (save-excursion
	(let ((header-end (line-to-char-offset (find-last-header-line)))
	      (attr nil)
	      (val nil)
	      (pairs nil))
	  (goto-line (find-first-header-line))
	  (while (and (< (point) header-end)
		      (re-search-forward
		       (format "^%s [A-Z][a-z]*:" header-comment)
		       header-end t))
	    (setq attr (upcase (buffer-substring (+ 2 (match-beginning 0))
						 (1- (match-end 0)))))
	    (goto-char (1+ (match-end 0)))
	    (if (re-search-forward "[^ 	].*$" (point-at-end-of-line) t)
		(setq val (buffer-substring (match-beginning 0) (match-end 0)))
	      (setq val nil))
	    (setq pairs (cons (cons attr val) pairs))
	    (forward-line 1)
	    )
	  pairs))))

(defun point-at-end-of-line ()
  "Returns the value of point at the end of the current line."
  (interactive)
  (set-buffer-header-vars)
  (save-excursion
    (end-of-line)
    (point)))

(provide 'header)
