;;; hide-ifdef-mode.el   Hides selected code within ifdefs.
;;;
;;; Copyright (C) 1988 Brian Marick and Daniel LaLiberte
;;; Written by Brian Marick, at Gould, Computer Systems Division, Urbana IL.
;;; Extensively modified by Daniel LaLiberte (while at Gould).
;;;
;;; You may freely modify and distribute this, but keep a record
;;; of modifications and send comments to:
;;; 	 liberte@a.cs.uiuc.edu  or  ihnp4!uiucdcs!liberte
;;; I will continue to upgrade hide-ifdef-mode
;;; with your contributions and will eventually offer it to FSF.
;;;
;;; $Header: hide-ifdef.e,v 2.1 89/04/24 18:37:50 darrylo Exp $
;;;
;;; Revision 1.3  88/01/29  00:38:19  liberte
;;; Fix three bugs.
;;; 1. Function "defined" is just like lookup.
;;; 2. Skip to newline or cr in case text is hidden.
;;; 3. Use car of token list if just one symbol.
;;; 
;;;
;;;--------------------------------------------------------------
;;; To initialize, toggle the hide-ifdef minor mode with
;;;
;;; M-x hide-ifdef-mode
;;;
;;; This will set up key bindings and call hide-ifdef-mode-hook if it
;;; has a value.  (The key binding business needs a little work.)
;;; To explicitly hide ifdefs using a buffer-local define list, type
;;;
;;; M-x hide-ifdefs  or C-c h
;;;
;;; Hide-ifdefs suppresses the display of code that the preprocessor wouldn't
;;; pass through.  The support of constant expressions in #if lines is 
;;; limited to identifiers, parens, and the operators && || ! "defined".
;;;
;;; The hidden code is marked by ellipses (...).  Be
;;; cautious when editing near ellipses, since the hidden text is
;;; still in the buffer, and you can move the point into it and modify
;;; text unawares.  If you don't want to see the ellipses, set 
;;; selective-display-ellipses to nil.  But this can be dangerous.
;;; You can make your buffer read-only while hiding by setting
;;; hide-ifdefs-read-only to a non-nil value.
;;; You can toggle this variable with toggle-hide-read-only (C-c C-q).
;;;
;;; You can undo the effect of hide-ifdefs by typing
;;;
;;; M-x show-ifdefs  or C-c s
;;;
;;; Use M-x define (C-c d) to define a variable.
;;; Use M-x undef (C-c u) to undefine a variable.
;;;
;;; If you Define or Undef a variable while Hide-Ifdefs is in effect,
;;; the display will be updated.  Only the define list for the current
;;; buffer will be affected.  You can save changes to the local define
;;; list with set-define-list.  This adds entries to a buffer local
;;; variable hide-define-alist.
;;;
;;; If you have defined a hide-ifdef-mode-hook, you can set
;;; up a list of symbols that may be used by hide-ifdefs as in the
;;; following example:
;;;
;;; (setq hide-ifdef-mode-hook
;;;      '(lambda ()
;;;	 (if (not hide-define-alist)
;;;	     (setq hide-define-alist
;;;		  '((list1 ONE TWO)
;;;		    (list2 TWO THREE)
;;;		    )))
;;;	 (use-define-list 'list2) ; use list2 by default
;;;	 ))
;;;
;;; You can call use-define-list (C-c u) at any time to specify
;;; another list to use.
;;;
;;; To cause ifdefs to be hidden as soon as hide-ifdefs-mode is called,
;;; set hide-ifdefs-initially to non-nil.
;;;
;;; If you set hide-ifdef-lines to t, hide-ifdefs hides all the #ifdef lines.
;;; In the absence of highlighting, that might be a bad idea.  If you set
;;; hide-ifdef-lines to nil (the default), the surrounding preprocessor
;;; lines will be displayed.  That can be confusing in its own
;;; right.  Other variations on display are possible, but not much
;;; better.
;;;
;;; You can explicitly hide or show individual ifdef blocks irrespective
;;; of the define list by using hide-ifdef-block and show-ifdef-block.
;;; You can move the point between ifdefs with forward-ifdef, backward-ifdef,
;;; up-ifdef, down-ifdef, next-ifdef, and previous-ifdef.
;;;
;;; The buffer-local variable 'hiding is true whenever text might be hidden.
;;; So you can display this state in your mode line with:
;;;
;;;   (setq minor-mode-alist (cons '(hiding " Hiding") minor-mode-alist))
;;;
;;; provided you have minor-mode-alist in your mode line (the default).
;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; From liberte@b.cs.uiuc.edu  Fri Jan 29 02:55:40 1988
;; Received: from b.cs.uiuc.edu by prep.ai.mit.edu; Fri, 29 Jan 88 02:55:40 EST
;; Received: by b.cs.uiuc.edu (UIUC-5.52/9.7)
;; 	id AA08632; Thu, 28 Jan 88 23:45:20 CST
;; Date: Thu, 28 Jan 88 23:45:20 CST
;; From: liberte@b.cs.uiuc.edu (Daniel LaLiberte)
;; Message-Id: <8801290545.AA08632@b.cs.uiuc.edu>
;; To: info-gnu-emacs@prep.ai.mit.edu
;; Subject: hide-ifdef-mode
;; 
;; Seems that lots of people would like to use this hide-ifdef-mode.
;; There is some documentation at the top of the file and in the
;; hide-ifdef-mode function.
;; 
;; Enjoy,
;; 
;; Daniel LaLiberte   217-333-7937
;; University of Illinois, Urbana-Champaign
;; Department of Computer Science
;; 1304 W Springfield
;; Urbana, IL  61801
;; 
;; liberte@a.cs.uiuc.edu
;; uiucdcs!liberte
;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;	Note: this version has had "official" patches applied to it.
;;	The patch header follows (this information is given for
;;	genealogical purposes).
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; From liberte@b.cs.uiuc.edu  Fri Jan 29 01:43:00 1988
;; Received: from b.cs.uiuc.edu by prep.ai.mit.edu; Fri, 29 Jan 88 01:43:00 EST
;; Received: by b.cs.uiuc.edu (UIUC-5.52/9.7)
;; 	id AA09418; Fri, 29 Jan 88 00:48:34 CST
;; Date: Fri, 29 Jan 88 00:48:34 CST
;; From: liberte@b.cs.uiuc.edu (Daniel LaLiberte)
;; Message-Id: <8801290648.AA09418@b.cs.uiuc.edu>
;; To: info-gnu-emacs@prep.ai.mit.edu
;; Subject: hide-ifdef-mode
;; 
;; Oops.  I sent out hide-ifdef-mode with some rather obvious bad bugs.
;; Last minute tinkering...
;; 
;; Apply these diffs please.
;; 
;; dan
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar hide-ifdef-mode-map nil
  "Keymap used with hide-ifdef mode")

(defconst  hide-ifdef-mode-prefix-key "\C-c"
  "Prefix key for all hide-ifdef-mode commands.")

(defvar before-hide-ifdef-mode-map nil
  "Buffer-local variable to store a copy of the local keymap
	before hide-ifdef-mode modifies it.")

(defun define-hide-ifdef-mode-map ()
  (if hide-ifdef-mode-map
      ()				; dont redefine it.
    (setq hide-ifdef-mode-map (make-sparse-keymap))
    (define-key hide-ifdef-mode-map "d" 'define)
    (define-key hide-ifdef-mode-map "u" 'undef)
    (define-key hide-ifdef-mode-map "D" 'set-define-list)
    (define-key hide-ifdef-mode-map "U" 'use-define-list)
  
    (define-key hide-ifdef-mode-map "h" 'hide-ifdefs)
    (define-key hide-ifdef-mode-map "s" 'show-ifdefs)
    (define-key hide-ifdef-mode-map "\C-h" 'hide-ifdef-block)
    (define-key hide-ifdef-mode-map "\C-s" 'show-ifdef-block)
  
    (define-key hide-ifdef-mode-map "\C-f" 'forward-ifdef)
    (define-key hide-ifdef-mode-map "\C-b" 'backward-ifdef)
    (define-key hide-ifdef-mode-map "\C-d" 'down-ifdef)
    (define-key hide-ifdef-mode-map "\C-u" 'up-ifdef)
    (define-key hide-ifdef-mode-map "\C-n" 'next-ifdef)
    (define-key hide-ifdef-mode-map "\C-p" 'previous-ifdef)
    (define-key hide-ifdef-mode-map "\C-q" 'toggle-hide-read-only)
    (define-key hide-ifdef-mode-map
      (where-is-internal 'toggle-read-only nil t) 'toggle-outside-read-only)
    )
  (fset 'hide-ifdef-mode-map hide-ifdef-mode-map)  ; the function is the map
  )

(defun update-mode-line ()
  "Update mode-line by setting buffer-modified to itself."
  (set-buffer-modified-p (buffer-modified-p)))

(defvar hide-ifdef-mode nil
  "non-nil when hide-ifdef-mode is activated.")
(make-variable-buffer-local 'hide-ifdef-mode)

(defun hide-ifdef-mode (arg)
  "Toggle hide-ifdef-mode.  Thus this is a minor mode, albeit a large one.
With arg, turn hide-ifdef-mode on iff arg is positive.
In hide-ifdef-mode, code within ifdefs that the C preprocessor
would eliminate may be hidden from view.  Several variables affect
ifdef hiding:

*hide-ifdef-env*
	An association list of defined and undefined symbols for the
	current buffer.
hide-define-alist
	An association list of defined symbol lists.  Use set-define-list
	to save the current *hide-ifdef-env* and use-define-list to set
	the current *hide-ifdef-env* from one of the lists in hide-define-alist.
hide-ifdef-lines
	Set to non-nil to not show #if, #ifdef, #ifndef, #else, and
	#endif lines when hiding.  Buffer local.
hide-ifdefs-initially
	Indicates whether ifdefs should be hidden on start up.
hide-ifdefs-read-only
	Set to non-nil if you want to make buffers read only while hiding.
	After show-ifdefs, read-only status is restored to previous value.

Initially, the global values of *hide-ifdef-env* and hide-ifdef-lines
are used.
\\{hide-ifdef-mode-map}"

  (interactive "P")
  (setq hide-ifdef-mode
	(if (null arg)
	    (not hide-ifdef-mode)
	  (> (prefix-numeric-value arg) 0)))
  
  (update-mode-line)

  (if hide-ifdef-mode
      (progn
	; fix c-mode syntax table so we can recognize whole symbols.
	(modify-syntax-entry ?_ "w")
	(modify-syntax-entry ?& ".")
	(modify-syntax-entry ?\| ".")

	; inherit global values
	(make-local-variable '*hide-ifdef-env*)
	(setq *hide-ifdef-env* (default-value '*hide-ifdef-env*))

	(make-local-variable 'hide-ifdef-lines)
	(setq hide-ifdef-lines (default-value 'hide-ifdef-lines))

	(make-local-variable 'hiding)
	(setq hiding (default-value 'hiding))

	(make-local-variable 'hide-outside-read-only)
	(setq hide-outside-read-only buffer-read-only)

	(make-local-variable 'before-hide-ifdef-mode-map)
	(setq before-hide-ifdef-mode-map
	      (current-local-map))
	(use-local-map (copy-keymap (current-local-map)))
	(local-unset-key hide-ifdef-mode-prefix-key)
	(local-set-key hide-ifdef-mode-prefix-key 'hide-ifdef-mode-map)
	(define-hide-ifdef-mode-map)

	(run-hooks 'hide-ifdef-mode-hook)

	(if hide-ifdefs-initially
	    (hide-ifdefs)
	  (show-ifdefs))
	(message "Enter hide-ifdef-mode.")
	)
     ; else end hide-ifdef-mode
    (if hiding
	(show-ifdefs))
    (use-local-map before-hide-ifdef-mode-map)
    (message "Exit hide-ifdef-mode.")
    ))
  

;; from outline.el with docstring fixed.
(defun outline-flag-region (from to flag)
  "Hides or shows lines from FROM to TO, according to FLAG.  If FLAG
is \\n (newline character) then text is shown, while if FLAG is \\^M
\(control-M) the text is hidden."
  (let ((modp (buffer-modified-p)))
    (unwind-protect (progn
		      (subst-char-in-region from to
			      (if (= flag ?\n) ?\^M ?\n)
			      flag t) )
      (set-buffer-modified-p modp))
    ))

(defun show-all ()
  "Show all of the text in the current buffer."
  (interactive)
  (outline-flag-region (point-min) (point-max) ?\n))

(defun hide-region (start end)
  "START is the start of a #if or #else form.  END is the ending part.
Everything including these lines is made invisible."
  (outline-flag-region start end ?\^M)
  )

(defun show-region (start end)
  "Everything between START and END is made visible."
  (outline-flag-region start end ?\n)
  )



;===%%SF%% evaluation (Start)  ===

(defvar *ifdef-evaluator* 'eval
  "The evaluator is given a canonical form and returns T if text under
that form should be displayed.")


(defvar *undefined-symbol-value* nil
  "...is by default considered to be false.")

(defvar *hide-ifdef-env* nil
  "An alist of defined symbols and their values.")

(defun set-var (var value)
  "Prepend (var value) pair to *hide-ifdef-env*."
  (setq *hide-ifdef-env* (cons (cons var value) *hide-ifdef-env*)))


(defun lookup (var)
;  (message "lookup %s" var)
  (let ((val (assoc var *hide-ifdef-env*)))
    (if val
	(cdr val)
      *undefined-symbol-value*)))

(defun defined (var)
  (lookup var)
)
;  (if (assoc var  *hide-ifdef-env*)
;      1
;    nil)


;===%%SF%% evaluation (End)  ===



;===%%SF%% parsing (Start)  ===
;;;  The code that understands what ifs and ifdefs in files look like.

(defconst cpp-prefix "^[ \t]*#[ \t]*")
(defconst ifndef-regexp (concat cpp-prefix "ifndef"))
(defconst ifx-regexp (concat cpp-prefix "if\\(n?def\\)?[ \t]+"))
(defconst else-regexp (concat cpp-prefix "else"))
(defconst endif-regexp (concat cpp-prefix "endif"))
(defconst ifelseendif-regexp
  (concat ifx-regexp "\\|" else-regexp "\\|" endif-regexp))


(defun infix-to-prefix (token-list)
  "Convert list of tokens in infix into prefix list"
;  (message "infix-to-prefix: %s" token-list)
  (if (= 1 (length token-list))
      (` (lookup (quote (, (car token-list)))))
    (hi-parse-if-exp token-list))
  )

; pattern to match initial identifier, !, &&, ||, (, or ).
(defconst token-regexp "^\\(!\\|&&\\|||\\|[()]\\|\\w+\\)")
(defconst end-of-comment "\\*/")


(defun tokenize (expr-string)
  "Separate string into a list of tokens"
  (let ((token-list nil)
	(expr-start 0)
	(expr-length (length expr-string)))

    (while (< expr-start expr-length) 
;      (message "expr-start = %d" expr-start) (sit-for 1)
      (cond
	((string-match "^\\([ \t]\\|\\\\n\\)+" expr-string expr-start)
	   ; skip whitespace
	 (setq expr-start (match-end 0))
	 ; stick newline in string so ^ matches on the next string-match
	 (aset expr-string (1- expr-start) ?\n)
	 )

	((string-match "^/\\*" expr-string expr-start)
	 (setq expr-start (match-end 0))
	 (aset expr-string (1- expr-start) ?\n)
	 (string-match end-of-comment expr-string expr-start) ; eat comment
	 (setq expr-start (match-end 0))
	 (aset expr-string (1- expr-start) ?\n)
	 )

	((string-match token-regexp expr-string expr-start)
	  (let ((token (substring expr-string expr-start (match-end 0))))
	    (setq expr-start (match-end 0))
	    (aset expr-string (1- expr-start) ?\n)
;	    (message "token: %s" token) (sit-for 1)
	    (setq token-list
		  (cons
		    (cond
		      ((string-equal token "||") 'or)
		      ((string-equal token "&&") 'and)
		      ((string-equal token "!")  'not)
		      ((string-equal token "defined") 'defined)
		      ((string-equal token "(") 'lparen)
		      ((string-equal token ")") 'rparen)
		      (t (intern token)))
		    token-list))
	    ))
	  (t (error "Bad #if expression: %s" expr-string))
	  ))
    (reverse token-list)
    ))

;;;-----------------------------------------------------------------
;;; Translate C preprocessor #if expressions using recursive descent.
;;; This parser is limited to the operators &&, ||, !, and "defined".

(defun hi-parse-if-exp (token-list)
  "Parse the TOKEN-LIST.  Return translated list in prefix form."
  (nexttoken)
  (prog1
      (hi-expr)
    (if token ; is there still a token?
	(error "Error: unexpected token: %s" token)))
  )

(defun nexttoken ()
  "Pop the next token from token-list into token."
  (setq token (car token-list))
  (setq token-list (cdr token-list))
  token
  )

(defun hi-expr ()
  (let ((result (hi-term)))
    (while (eq  token 'or)
      (nexttoken)
      (setq result (list 'or result (hi-term))))
  result
  ))

(defun hi-term ()
  (let ((result (hi-factor)))
    (while (eq token 'and)
      (nexttoken)
      (setq result (list 'and result (hi-factor))))
    result
    ))

(defun hi-factor ()
  (cond
    ((eq token 'not)
     (nexttoken)
     (list 'not (hi-factor)))

    ((eq token 'lparen)
     (nexttoken)
     (let ((result (hi-expr)))
       (if (not (eq token 'rparen))
	   (error "Bad token in parenthesized expression: %s" token)
	 (nexttoken)
	 result)))

    ((eq token 'defined)
     (nexttoken)
     (if (not (eq token 'lparen))
	 (error "Error: expected \"(\" after \"define\""))
     (nexttoken)
     (let ((ident token))
       (if (memq token '(or and not defined lparen rparen))
	   (error "Error: unexpected token: %s" token))
       (nexttoken)
       (if (not (eq token 'rparen))
	   (error "Error: expected \")\" after identifier"))
       (nexttoken)
       (` (defined (quote (, ident))))
       ))

    (t ; identifier
      (let ((ident token))
	(if (memq ident '(or and))
	    (error "Error: missing identifier"))
	(nexttoken)
	(` (lookup (quote (, ident))))
	))

    ))

;;;----------- end of parser -----------------------


(defun canonicalize ()
  "When at beginning of #ifX, returns a canonical (evaluable)
       form for the expression."
  (save-excursion
    (let ((negate (looking-at ifndef-regexp)))
      (re-search-forward ifx-regexp)
      (let* ((expr-string
	      (buffer-substring (point)
				(progn (skip-chars-forward "^\n\r") (point))))
	     (expr (infix-to-prefix (tokenize expr-string))))
;	(message "canonicalized: %s" expr)
	(if negate
	    (list 'not expr)
	  expr)))))


(defun find-any-ifX ()
  "Position at beginning of next #if, #ifdef, or #ifndef, including one on
this line."
;  (message "find ifX at %d" (point))
  (prog1
      (re-search-forward ifx-regexp (point-max) t)
    (beginning-of-line)))


(defun find-next-relevant ()
  "Position at beginning of next #ifdef, #ifndef, #else, #endif,
NOT including one on this line."
;  (message "find-next-relevant at %d" (point))
  (end-of-line)
  ; avoid infinite recursion by only going to beginning of line if match found
  (if (re-search-forward ifelseendif-regexp (point-max) t)
      (beginning-of-line))
  )

(defun find-previous-relevant ()
  "Position at beginning of previous #ifdef, #ifndef, #else, #endif,
NOT including one on this line."
;  (message "find-previous-relevant at %d" (point))
  (beginning-of-line)
  ; avoid infinite recursion by only going to beginning of line if match found
  (if (re-search-backward ifelseendif-regexp (point-min) t)
     (beginning-of-line)
    )
  )


(defun looking-at-ifX ()		;; Should eventually see #if
  (looking-at ifx-regexp))
(defun looking-at-endif ()
  (looking-at endif-regexp))
(defun looking-at-else ()
  (looking-at else-regexp))



(defun ifdef-to-endif ()
  "If positioned at #ifX form, skips to corresponding #endif."
;  (message "ifdef-to-endif at %d" (point)) (sit-for 1)
  (find-next-relevant)
  (cond ((looking-at-ifX)
	 (ifdef-to-endif) ; find endif of nested if
	 (ifdef-to-endif)) ; find outer endif or else
	((looking-at-else)
	 (ifdef-to-endif)) ; find endif following else
	((looking-at-endif)
	 'done)
	(t
	 (error "Missmatched #ifdef #endif pair"))
	))


(defun endif-to-ifdef ()
  "If positioned at #endif form, skips backward to corresponding #ifX."
;  (message "endif-to-ifdef at %d" (point))
  (let ((start (point)))
    (find-previous-relevant)
    (if (= start (point))
	(error "Missmatched #ifdef #endif pair")))
  (cond ((looking-at-endif)
	 (endif-to-ifdef) ; find beginning of nested if
	 (endif-to-ifdef)) ; find beginning of outer if or else
	((looking-at-else)
	 (endif-to-ifdef))
	((looking-at-ifX)
	 'done)
	(t ; never gets here
	 )))


(defun forward-ifdef ()
  "Move point to beginning of line of the next ifdef-endif."
  (interactive)
  (let ((start (point)))
    (if (not (looking-at-ifX))
	  (find-next-relevant))
    (if (looking-at-ifX)
	(ifdef-to-endif)
      (goto-char start)
      (error "No following #ifdefs")
      )))


(defun backward-ifdef ()
  "Move point to beginning of previous ifdef-endif."
  (interactive)
  (beginning-of-line)
  (let ((start (point)))
    (if (not (looking-at-endif))
	(find-previous-relevant))
    (if (looking-at-endif)
	(endif-to-ifdef)
      (goto-char start)
      (error "No previous #ifdefs")
      )))



(defun down-ifdef ()
  "Move point to beginning of nested ifdef or else-part."
    (interactive)
    (let ((start (point)))
      (find-next-relevant)
      (if (or (looking-at-ifX) (looking-at-else))
	  ()
	(goto-char start)
	(error "No following #ifdefs")
	)))


(defun up-ifdef ()
  "Move point to beginning of enclosing ifdef or else-part."
  (interactive)
  (beginning-of-line)
  (let ((start (point)))
    (if (not (looking-at-endif))
	(find-previous-relevant))
    (if (looking-at-endif)
	(endif-to-ifdef))
      (if (= start (point))
	  (error "No previous #ifdefs")
	)))

(defun next-ifdef ()
  "Move to the beginning of the next #ifX, #else, or #endif"
  (interactive)
  (find-next-relevant)
  (if (eolp)
      (progn
	(beginning-of-line)
	(error "No following #ifdefs, #elses or #endifs")
	)))

(defun previous-ifdef ()
  "Move to the beginning of the previous #ifX, #else, or #endif"
  (interactive)
  (let ((start (point)))
    (find-previous-relevant)
    (if (= start (point))
	(error "No previous #ifdefs, #elses or #endifs")
      )))


;===%%SF%% parsing (End)  ===


;===%%SF%% hiding (Start)  ===


;;; A range is a structure with four components:
;;; ELSE-P	True if there was an else clause for the ifdef.
;;; START	The start of the range. (beginning of line)
;;; ELSE	The else marker (beginning of line)
;;;			Only valid if ELSE-P is true.
;;; END		The end of the range.  (beginning of line)

(defun make-range (else-p start end &optional else)
  (list else-p start else end))

(defun range-else-p (range)  (elt range 0))
(defun range-start (range) (elt range 1))
(defun range-else (range) (elt range 2))
(defun range-end (range) (elt range 3))



;;; Find-Range
;;; The workhorse, it delimits the #if region.  Reasonably simple:
;;; Skip until an #else or #endif is found, remembering positions.  If
;;; an #else was found, skip some more, looking for the true #endif.

(defun find-range ()
  "Returns a Range structure describing the current #if region.
Point is left unchanged."
;  (message "find-range at %d" (point))
  (save-excursion
    (beginning-of-line)
    (let ((start (point))
	  (else-p nil)
	  (else nil)
	  (end nil))
      ;; Part one.  Look for either #endif or #else.
      ;; This loop-and-a-half dedicated to E. Dijkstra.
      (find-next-relevant)
      (while (looking-at-ifX)		; Skip nested ifdef
	(ifdef-to-endif)
	(find-next-relevant))
      ;; Found either a #else or an #endif.
      (cond ((looking-at-else)
	     (setq else-p t)
	     (setq else (point)))
	    (t
	     (setq end (point)) ; (save-excursion (end-of-line) (point))
	     ))
      ;; If found #else, look for #endif.
      (if else-p
	  (progn
	    (find-next-relevant)
	    (while (looking-at-ifX)	; Skip nested ifdef
	      (ifdef-to-endif)
	      (find-next-relevant))
	    (if (looking-at-else)
		(error "Found two elses in a row?  Broken!"))
	    (setq end (point))  ; (save-excursion (end-of-line) (point))
	    ))
      (make-range else-p start end else))))

	  
;;; A bit slimy.
;;; NOTE:  If there's an #ifdef at the beginning of the file, we can't
;;; hide it.  There's no previous newline to replace.  If we added
;;; one, we'd throw off all the counts.  Feh.

(defun hide-line (point)
  "Hide the line containing point.  Does nothing if
hide-ifdef-lines is nil."
  (if hide-ifdef-lines
      (save-excursion
	(goto-char point)
	(let ((modp (buffer-modified-p)))
	  (unwind-protect
	      (progn
		(beginning-of-line)
		(if (not (= (point) 1))
		    (hide-region (1- (point)) (point))))
	    (set-buffer-modified-p modp))
	  ))
    ))
		  

;;;  Possibly-Hide
;;;  There are four cases.  The #ifX expression is "taken" if it
;;;  the *ifdef-evaluator* returns T.  Presumably, this means the code
;;;  inside the #ifdef would be included when the program was
;;;  compiled.  
;;;
;;;  Case 1:  #ifX taken, and there's an #else.
;;;	The #else part must be hidden.  The #if (then) part must be
;;;	processed for nested #ifX's.
;;;  Case 2:  #ifX taken, and there's no #else.
;;;	The #if part must be processed for nested #ifX's.
;;;  Case 3:  #ifX not taken, and there's an #else.
;;;	The #if part must be hidden.  The #else part must be processed
;;;	for nested #ifs.
;;;  Case 4:  #ifX not taken, and there's no #else.
;;;	The #ifX part must be hidden.
;;;
;;;  Further processing is done by narrowing to the relevant region
;;;  and just recursively calling hide-ifdefs-guts.
;;;
;;;  When possibly-hide returns, point is at the end of the
;;;  possibly-hidden range.

(defun recurse-on (start end)
  "Call hide-ifdefs-guts after narrowing to end of START line and END
line."
  (save-excursion
    (save-restriction
      (goto-char start)
      (end-of-line)
      (narrow-to-region (point) end)
      (hide-ifdefs-guts))))

(defun possibly-hide ()
  "Called at #ifX expression, this hides those parts that should be
hidden, according to judgement of *ifdef-evaluator*."
;  (message "possibly-hide") (sit-for 1)
    (let ((test (canonicalize))
	  (range (find-range)))
;      (message "test = %s" test) (sit-for 1)
      
      (hide-line (range-end range))
      (if (funcall *ifdef-evaluator* test)
	  (cond ((range-else-p range) ; case 1
		 (hide-line (range-else range))
		 (hide-region (range-else range) (1- (range-end range)))
		 (recurse-on (range-start range) (range-else range)))
		(t ; case 2
		 (recurse-on (range-start range) (range-end range))))
	(cond ((range-else-p range) ; case 3
	       (hide-line (range-else range))
	       (hide-region (range-start range) (1- (range-else range)))
	       (recurse-on (range-else range) (range-end range)))
	      (t ; case 4
	       (hide-region (point) (1- (range-end range))))
	      ))
      (hide-line (range-start range))	; Always hide start.
      (goto-char (range-end range))
      (end-of-line)
      ))



(defun hide-ifdefs-guts ()
  "Does the work of hide-ifdefs, except for the work that's pointless
to redo on a recursive entry."
;  (message "hide-ifdefs-guts")
  (save-excursion
    (goto-char (point-min))
    (while (find-any-ifX)
      (possibly-hide))))

;===%%SF%% hiding (End)  ===


;===%%SF%% exports (Start)  ===

(defvar hide-ifdefs-initially nil
  "*Non-nil if hide-ifdefs should be called when hide-ifdef-mode is activated.")

(defvar hiding nil
  "Tells whether we're hiding or not.")

(defvar hide-ifdefs-read-only nil
  "*Set to non-nil if you want buffer to be read-only while hiding text.")

(defvar hide-outside-read-only nil
  "Internal variable.  Saves the value of buffer-read-only while hiding.")

(defvar hide-ifdef-lines nil
  "*Set to t if you don't want to see the #ifX, #else, and #endif lines.")

(defun toggle-hide-read-only ()
  "Toggle hide-ifdefs-read-only."
  (interactive)
  (setq hide-ifdefs-read-only (not hide-ifdefs-read-only))
  (message "Hide-Read-Only %s" (if hide-ifdefs-read-only "ON" "OFF"))
  (if hiding
      (setq buffer-read-only (or hide-ifdefs-read-only hide-outside-read-only)))
  (update-mode-line)
  )

(defun toggle-outside-read-only ()
  "Replacement for toggle-read-only in hide-ifdef-mode."
  (interactive)
  (setq hide-outside-read-only (not hide-outside-read-only))
  (message "Read only %s" (if hide-outside-read-only "ON" "OFF"))
  (setq buffer-read-only
	(or (and hiding hide-ifdefs-read-only)
	    hide-outside-read-only)
	)
  (update-mode-line)
  )

      
(defun define (var)
  "Define a VAR so that #ifdef VAR would be included."
  (interactive "SDefine what? ")
  (set-var var t)
  (if hiding (hide-ifdefs)))

(defun undef (var)
  "Undefine a VAR so that #ifdef VAR would not be included."
  (interactive "SUndefine what? ")
  (set-var var nil)
  (if hiding (hide-ifdefs)))


(defun hide-ifdefs ()
  "Hides certain text.  Assume *hide-ifdef-env* has been set up with
calls to define and undef.  The text hidden is the text that would not
be included by the C preprocessor if it were given the file with those
symbols defined or undefined.

Turn off by calling show-ifdefs."

  (interactive)
  (message "Hiding...")
  (if (not hide-ifdef-mode)
      (hide-ifdef-mode 1)) ; turn on hide-ifdef-mode
  (if hiding
      (show-ifdefs))			; Otherwise, deep confusion.
  (if buffer-read-only (toggle-read-only)) ; make it writable temporarily
  (setq selective-display t)
  (setq hiding t)
  (hide-ifdefs-guts)
  (if (or hide-ifdefs-read-only hide-outside-read-only)
      (toggle-read-only) ; make it read only
    )
  (message "Hiding done")
  )


(defun show-ifdefs ()
  "Cancel the effects of hide-ifdefs."
  (interactive)
;  (message "show-ifdefs") (sit-for 1)
  (if buffer-read-only (toggle-read-only)) ; make it writable temporarily
  (setq selective-display nil)	; defaults
  (show-all)
  (if hide-outside-read-only
      (toggle-read-only)) ; make it read only
  (setq hiding nil)
  )


(defun find-ifdef-block ()
  "Utilitiy for hide and show ifdef-block.  Set top and bottom of ifdef block."
  (save-excursion
    (beginning-of-line)
    (if (not (or (looking-at-else) (looking-at-ifX)))
	(up-ifdef))
    (setq top (point)))
  (save-excursion
    (beginning-of-line)
    (if (not (looking-at-endif))
	(find-next-relevant))
    (while (looking-at-ifX)
      (ifdef-to-endif)
      (find-next-relevant)
      )
    (setq bottom (1- (point))))
  )


(defun hide-ifdef-block ()
  "Hide the ifdef block (true or false part) enclosing the cursor."
  (interactive)
  (if (not hide-ifdef-mode)
      (hide-ifdef-mode 1))
  (if buffer-read-only (toggle-read-only))
  (setq selective-display t)
  (let (top bottom)
    (find-ifdef-block) ; set top and bottom - dynamic scoping
    (hide-region top bottom)
    (if hide-ifdef-lines
	(progn
	  (hide-line top)
	  (hide-line (1+ bottom))))
    (setq hiding t)
    )
  (if (or hide-ifdefs-read-only hide-outside-read-only)
      (toggle-read-only))
  )


(defun show-ifdef-block ()
  "Show the ifdef block (true or false part) enclosing the cursor."
  (interactive)
  (let ((old-read-only buffer-read-only))
    (if old-read-only (toggle-read-only))
    (if hide-ifdef-lines
	(save-excursion
	  (beginning-of-line)
	  (show-region (1- (point)) (progn (end-of-line) (point))))

      (let (top bottom)
	(find-ifdef-block)
	(show-region (1- top) bottom))
      )

    ; leave read only status the same since we dont know if all is shown.
    (if old-read-only (toggle-read-only))
    ))

(defvar hide-define-alist nil
  "A global assoc list of predefined *hide-ifdef-env* lists")

(defun compress-define-list (env)
  "Compress the define list ENV into a list of defined symbols."
  (let ((defs (mapcar '(lambda (arg) (if (lookup (car arg)) (car arg)))
		      env))
	(new-defs nil))
    (while defs
      (if (car defs)
	  (setq new-defs (cons (car defs) new-defs)))
      (setq defs (cdr defs)))
    new-defs
    ))

(defun set-define-list (name)
  "Set the association for NAME to *hide-ifdef-env*"
  (interactive "SSet define list:")
  (setq hide-define-alist
	(cons (cons name (compress-define-list *hide-ifdef-env*))
	      hide-define-alist))
  )

(defun use-define-list (name)
  "Set *hide-ifdef-env* to the define list specified by NAME"
  (interactive "SUse define list:")
  (let ((define-list (assoc name hide-define-alist)))
    (if define-list
	(setq *hide-ifdef-env*
	      (mapcar '(lambda (arg) (cons arg t))
		      (cdr define-list)))
      (error "No define list for %s" name))
    (if hiding (hide-ifdefs))
    )
  )

;===%%SF%% exports (End)  ===

