(defun midas-mode ()
  "Major mode for editing assembly language programs."
  (interactive)
  (kill-all-local-variables)
  (use-local-map midas-mode-map)
  (setq major-mode 'midas-mode)
  (setq mode-name "Midas")
  (set-syntax-table midas-mode-syntax-table)
  (setq local-abbrev-table midas-mode-abbrev-table)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'insert-tab)
  (make-local-variable 'indent-tabs-mode)
  (setq indent-tabs-mode t)
  (make-local-variable 'comment-start)
  (setq comment-start "# ")
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "#+ *")
  (make-local-variable 'comment-column)
  (setq comment-column 40)
  (make-local-variable 'comment-indent-hook)
  (setq comment-indent-hook 'midas-comment-indent)
  (run-hooks 'midas-mode-hook))

(defvar midas-mode-syntax-table nil "")
(defvar midas-mode-abbrev-table nil "")
(defvar midas-mode-map (make-sparse-keymap) "")

(if (not midas-mode-syntax-table)
    (progn (setq midas-mode-syntax-table (make-syntax-table))
	   (set-syntax-table midas-mode-syntax-table)
	   (modify-syntax-entry ?\# "<   ")
	   (modify-syntax-entry ?\n ">   ")
	   (modify-syntax-entry ?. "w   ") ;e.g. 68000 move.w
	   (modify-syntax-entry ?% "_   ") ;e.g. %sp
	   (modify-syntax-entry ?$ "_   ") ;e.g. $1 for m4 preprocessor
	   (modify-syntax-entry ?` "('  ") ;for m4 preprocessor.
	   (modify-syntax-entry ?' ")`  ")
	   (modify-syntax-entry ?& "'   ")
	   ))

(define-abbrev-table 'midas-mode-abbrev-table ())

(defun midas-comment-indent (&optional pos)
  (save-excursion
    (if pos (goto-char pos))
    (cond ((looking-at "###") (current-column))
	  ((looking-at "##") 8)
	  (t
	   (skip-chars-backward " \t")
	   (max (if (bolp) 0 (1+ (current-column)))
		comment-column)))))