; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         next-buffer.el
; RCS:          $Header: next-buffer.,v 2.1 89/04/01 20:57:35 darrylo Exp $
; Description:  In the current window, switch to editing the next buffer
;		in the buffer list.  A list of buffers to skip over are
;		contained in the variable next-buffers to skip.
; Author:       Darryl Okahata
; Created:      Thu Mar 30 11:24:44 1989
; Modified:     Sat Feb 23 17:13:47 1991 (Darryl Okahata)
; Language:     Emacs-Lisp
; Package:      N/A
;
; (C) Copyright 1987, 1989, 1990, 1991, Hewlett-Packard, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;*****************************************************************************
;;
;;	next-buffer
;;
;;	In the current window, switch to editing the next buffer in the buffer
;;	list.  A list of buffers to skip over are contained in the variable
;;	next-buffers-to-skip.
;;
;;	Programmer:	Darryl Okahata
;;
;;*****************************************************************************
;;
;;	Original:	August 30, 1987
;;

(defvar next-buffers-to-skip (list
			      "^[ ]*\*Minibuf-"
			      "^[ ]*\*Completions\*"
			      "^[ ]*\*Occur menu\*"
			      "^[ ]*\*Help\*"
			      "^[ ]*\*Buffer List\*"
			      )
 "List of buffers to skip when moving to the next buffer \[next-buffer]")


(defun next-buffer ()
  "In the current window, switch to editing the next buffer in the buffer list.
A list (in regexp form) of buffers to skip over are contained in the variable
next-buffers-to-skip."
  (interactive "")
  (let ( (index (next-buffer-find-current-index)) (buf-list (buffer-list)) )
    					; Move to next buffer
    (setq index (next-buffer-next-buffer index))
    					; While the new buffer is one we should
    					;    skip ...
    (while (next-buffer-skip (buffer-name (nth index buf-list)))
      (setq index (next-buffer-next-buffer index))
    )
    (switch-to-buffer (nth index buf-list) t)
  )
)


(defun next-buffer-next-buffer (index)
  (let ( (blen (length (buffer-list))) )
    (setq index (+ index 1))		; Move to next buffer
    (cond ( (<= blen index) (setq index 0) ) )	; Wrap around if necessary
  )
  index
)


;------------------------------------------------------------------------------
;
;	next-buffer-skip
;
;	Check the buffer 'buffer' to see if it is one we should skip.
;	Return t if it is, else return nil.
;

(defun next-buffer-skip (buffer)
  (let ( (skip-list next-buffers-to-skip) (done nil) (skip-buffer nil)
	item-to-skip )
    (while (not done)
      (setq item-to-skip (car skip-list))	; Get item to skip
      (setq skip-list (cdr skip-list))		; Pop it off the list
      (cond
	( (string-match buffer item-to-skip)	; If we should skip it ...
	      (progn
		(setq skip-buffer t)
	        (setq done t)
	      )
	)
	( (not skip-list) (setq done t))
      )
    )
    skip-buffer
  )
)


;------------------------------------------------------------------------------
;
;	next-buffer-find-current-index
;
;	Return the index (into the buffer list) of the currently displayed
;	buffer.
;

(defun next-buffer-find-current-index ()
  (let ( (current-buffer (buffer-name)) (blist (buffer-list)) (index 0) )
    (while (not (string-equal current-buffer (buffer-name (nth index blist))))
      (setq index (+ index 1))
    )
    index
  )
)

