; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         objc.el
; RCS:          $Header: $
; Description:  Objective-C helper routines
; Author:       Robert Nicholson
; Created:      Sun May 22 12:03:10 1994
; Modified:     Sun May 22 17:08:44 1994 (Robert Nicholson) robert@steffi.demon.co.uk
; Language:     Emacs-Lisp
; Package:      N/A
; Status:       Experimental (Do Not Distribute)
;
; (C) Copyright 1994, Robert Nicholson, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun objc-subclass-region (reg-start reg-end)
  "Build interface and implementation skeletons for subclass of region"
  (interactive "r")
  (let* ((subclassname (read-string "Name of subclass: "))
	 (superclassname (buffer-substring reg-start reg-end))
	 (subclassheadername (concat subclassname ".h"))
	 (superclassheadername (concat superclassname ".h")))
    (find-file-other-window subclassheadername)
    (objc-create-subclass-interface subclassname superclassname superclassheadername)
    (find-file-other-window (concat subclassname ".m"))
    (objc-create-subclass-implementation subclassheadername subclassname)))
   
(defun objc-create-subclass-interface (subclassname superclassname superclassheadername)
  "Build interface skeleton for new subclass"
    (insert (format "#import %s\n\n" superclassheadername))
    (insert (format "@interface %s:%s\n" subclassname superclassname))
    (insert "{\n}\n")
    (insert "@end\n"))

(defun objc-create-subclass-implementation (subclassheadername subclassname)
  "Build implementation skeleton for new subclass"
  (insert (format "#import %s\n\n" subclassheadername))
  (insert (format "@implementation %s\n" subclassname))
  (insert "@end\n"))

(defun objc-category-region (reg-start reg-end)
  "Build interface and implementation skeletons for category of region"
  (interactive "r")
   (let* ((categoryname (read-string "Name of category: "))
	 (classname (buffer-substring reg-start reg-end))
	 (classheadername (concat classname ".h"))
	 (categoryheadername (format "%s.%s.h" classname categoryname))
	 (categoryimplementationname (format "%s.%s.m" classname categoryname)))
     (find-file-other-window categoryheadername)
     (objc-create-category-interface categoryname classname classheadername)
     (find-file-other-window categoryimplementationname)
     (objc-create-category-implementation categoryname classname categoryheadername)))

(defun objc-create-category-interface (categoryname classname classheadername)
  "Build interface skeleton for new category"
  (insert (format "#import %s\n\n" classheadername))
  (insert (format "@interface %s(%s)\n" classname categoryname))
  (insert "@end\n"))

(defun objc-create-category-implementation (categoryname classname categoryheadername)
  "Build implementation skeleton for new category"
  (insert (format "#import %s\n\n" categoryheadername))
  (insert (format "@implementation %s(%s)\n" classname categoryname))
  (insert "@end\n"))
