;******************************************************************************
;* Customization of GNU-emacs * Per Abrahamsen * 1987 * amanda@iesd * auc * X *
;******************************************************************************
;; outline-minor-mode is a minor-mode version of outline-mode.
;; 
;; To use, insert outline-m.el and minor-map.el somewhere in your emacs
;; load-path, eg. under ../src/emacs/lisp, and insert init.el in your
;; .emacs file or in your site-init.el.
;; 
;; The command "M-x outline-minor-mode" will then add outline capability
;; to the current major mode. Per default, outline levels is controled by
;; the regular expression "[*\^l]+" and the length of the matching
;; string. This can be changed on a per major mode basis, and init.el
;; contain examples of hooks for use with latex-mode and emacs-lisp-mode.
;; 
;; outline-minor-mode must use a prefix key for its commands. As C-c is
;; reserved to the major mode, it use C-l You can change this with by
;; setting outline-prefix-char in your .emacs.
;; Type "C-h f outline-minor-mode" to see the commands.
;; 
;; GNU Emacs don't really allow minor modes to change the keymap, so I
;; wrote minor-map in an attemp to overcome this. At the time, I thought
;; I would be writing lots of minor modes, and that they would compete
;; for the keys. It is an overkill just for outline-minor-mode, but it
;; works, and I am to lazy to change it.
;; 
;; Bugs, comments etc. to 
;; 
;; Per Abrahamsen, alias amanda@iesd.dk.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; NOTE: init.el has been renamed to out-init.el, and minor-map.el has been
;; renamed to out-map.el.
;;

;******************************************************************************
;* AutoLoad                                                                   *
;******************************************************************************

; (setq load-path (cons "/home/mdc/amanda/lisp" load-path))

(autoload 'outline-minor-mode "outline-m"
	  "Show outline of current buffer." t)

;******************************************************************************
;* Outline Hooks                                                              *
;******************************************************************************

(setq emacs-lisp-mode-hook		;Use of OutLine
      '(lambda ()
	 (setq outline-regexp "(def")))

(defvar LaTeX-outline-regexp "[ \t]*\\\\\\(appendix\\|documentstyle\\|part\\|chapter\\|section\\|subsection\\|subsubsection\\|paragraph\\|subparagraph\\)")

(defun LaTeX-outline-level ()
  "Find the level of current outline heading in an LaTeX document."
  (save-excursion
    (skip-chars-forward " \t")
    (forward-char 1)
    (cond ((looking-at "subparagraph") 9)
	  ((looking-at "paragraph") 8)
	  ((looking-at "subsubsection") 7)
	  ((looking-at "subsection") 6)
	  ((looking-at "section") 5)
	  ((looking-at "chapter") 4)
	  ((looking-at "part") 3)
	  ((looking-at "appendix") 2)
	  ((looking-at "documentstyle") 2))))

(setq LaTeX-mode-hook			;Use of Outline
      '(lambda ()
	 (setq outline-regexp LaTeX-outline-regexp)
	 (setq outline-level-function 'LaTeX-outline-level)))

;******************************************************************************
;* Defaults                                                                   *
;******************************************************************************

; There have to be a better way.

(make-variable-buffer-local 'outline-prefix-char)
(setq-default outline-prefix-char "\C-l")

(make-variable-buffer-local 'outline-regexp)
(setq-default outline-regexp "[*\^l]+")

(make-variable-buffer-local 'outline-level-function)
(setq-default outline-level-function 'outline-level-default)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; From mcvax!kth!sunic!slyrf!dkuug!iesd!amanda@uunet.uu.net Fri Apr  7 14:19:26 1989
;; Date: 30 Mar 89 14:01:45 GMT
;; From: mcvax!kth!sunic!slyrf!dkuug!iesd!amanda@uunet.uu.net  (Per Abrahamsen)
;; Organization: Games Research, University of Aalborg, Denmark
;; Subject: Re: outline mode in emacs
;; References: <7708@boulder.Colorado.EDU>, <1555@iesd.dk>, <13484@steinmetz.ge.com>
;; Sender: info-gnu-emacs-request@prep.ai.mit.edu
;; To: info-gnu-emacs@prep.ai.mit.edu
;; 
;; 
;; In article <13484@steinmetz.ge.com> blackje@sunspot.steinmetz (Emmett Black) writes:
;; 
;; > How does it differ from the outline mode already included with GNU emacs?
;; 
;; It is just a minor mode version of the standard outline-mode. Being a
;; minor mode, it allows you to use outlines in other major modes, like
;; emacs-lisp-mode and latex-mode.
;; 
;; It also allow you to change the way outline levels is determinated in
;; these modes, by setting outline-regexp and outline-level-function in
;; the major mode hooks.
;; 
;; > Please describe it for us, and possibly consider posting it.
;; 
;; It follows after my signature...
;; 
;; Per Abrahamsen,  amanda@iesd.dk, {...}!mcvax!diku!iesd!amanda
;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
