;; Minor mode keymap for Emacs
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; outline-minor-mode is a minor-mode version of outline-mode.
;; 
;; To use, insert outline-m.el and minor-map.el somewhere in your emacs
;; load-path, eg. under ../src/emacs/lisp, and insert init.el in your
;; .emacs file or in your site-init.el.
;; 
;; The command "M-x outline-minor-mode" will then add outline capability
;; to the current major mode. Per default, outline levels is controled by
;; the regular expression "[*\^l]+" and the length of the matching
;; string. This can be changed on a per major mode basis, and init.el
;; contain examples of hooks for use with latex-mode and emacs-lisp-mode.
;; 
;; outline-minor-mode must use a prefix key for its commands. As C-c is
;; reserved to the major mode, it use C-l You can change this with by
;; setting outline-prefix-char in your .emacs.
;; Type "C-h f outline-minor-mode" to see the commands.
;; 
;; GNU Emacs don't really allow minor modes to change the keymap, so I
;; wrote minor-map in an attemp to overcome this. At the time, I thought
;; I would be writing lots of minor modes, and that they would compete
;; for the keys. It is an overkill just for outline-minor-mode, but it
;; works, and I am to lazy to change it.
;; 
;; Bugs, comments etc. to 
;; 
;; Per Abrahamsen, alias amanda@iesd.dk.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; NOTE: init.el has been renamed to out-init.el, and minor-map.el has been
;; renamed to out-map.el.
;;

;; Use minor-set-key to set a key in the minor keymap for a given mode.
;; Use unbind-minor-mode to remove all keydefinitions for a given
;; minor mode

;; COPYLEFT

;; Created 1987 by Per Abrahamsen at University of Aalborg, Denmark.

;; This file is not part of GNU emacs.

;; Everyone is granted permission to copy, modify and redistribute
;; this file, this include the right to remove this message.

;; HOW MINOR-MAP WORKS!

;; GNU Emacs minor modes may not alter the keymap. This is why modes
;; like outline-mode and indented-text-mode is major modes.
;; Minor modes could of cause just change the local keybinding on entry,
;; and then restore it on exit. But this will fail, if some other
;; minor mode change the keybinding meanwhile.

;; The minor keymap is contained in a local variable "minor-keymap"
;; descriped in the documentation of "make-minor-keymap".
;; When minor-set-key is called the first time with some key, a new
;; entry in minor-keymap with mode-list containing 'major-mode and
;; it's definition, and the name of the minor mode with definition.
;; The local binding of key is set to the definition used by the minor
;; mode.

;; Later call to minor-set-key will store the name and definition at
;; the head of the mode-list, and bind the key to definition. When a
;; minor mode is turned of, it should call unbind-minor-mode. It will
;; remove their entries in all the mode-lists. If the entry is at the
;; head of a mode-list, the key will be bound to the definition part
;; of the next entry.

;; RESTRICTIONS

;; The minor maps might fail, if the major mode (or the user) change
;; the local key bindings after the first call to minor-set-key.

;; Minor maps should support shared keymaps for prefix keys.

;; DOES MINOR-MAP WORK?

;; Please report any bugs (or a formal prove for correctness) to
;; amanda@iesd.uucp

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; From mcvax!kth!sunic!slyrf!dkuug!iesd!amanda@uunet.uu.net Fri Apr  7 14:19:26 1989
;; Date: 30 Mar 89 14:01:45 GMT
;; From: mcvax!kth!sunic!slyrf!dkuug!iesd!amanda@uunet.uu.net  (Per Abrahamsen)
;; Organization: Games Research, University of Aalborg, Denmark
;; Subject: Re: outline mode in emacs
;; References: <7708@boulder.Colorado.EDU>, <1555@iesd.dk>, <13484@steinmetz.ge.com>
;; Sender: info-gnu-emacs-request@prep.ai.mit.edu
;; To: info-gnu-emacs@prep.ai.mit.edu
;; 
;; 
;; In article <13484@steinmetz.ge.com> blackje@sunspot.steinmetz (Emmett Black) writes:
;; 
;; > How does it differ from the outline mode already included with GNU emacs?
;; 
;; It is just a minor mode version of the standard outline-mode. Being a
;; minor mode, it allows you to use outlines in other major modes, like
;; emacs-lisp-mode and latex-mode.
;; 
;; It also allow you to change the way outline levels is determinated in
;; these modes, by setting outline-regexp and outline-level-function in
;; the major mode hooks.
;; 
;; > Please describe it for us, and possibly consider posting it.
;; 
;; It follows after my signature...
;; 
;; Per Abrahamsen,  amanda@iesd.dk, {...}!mcvax!diku!iesd!amanda
;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'out-map)

(make-variable-buffer-local 'minor-keymap) ;Documentation?
(setq-default minor-keymap nil)

(make-variable-buffer-local 'minor-local-keymap)
(setq-default minor-local-keymap nil)

(defun make-minor-keymap ()
  "Construct a keymap for minor modes if nonexisting.

The keymap is an alist of (KEY . MODE-LIST) where MODE-LIST is a list
of (NAME . COMMAND).

Each KEY forms an entry to the alist. If a key sequence have no
entry, it mean that the KEY is bound to the default value in the
major mode map.

NAME identify a minor mode, that have bound KEY to COMMAND.
The head of MODE-LIST represent the newest, and currently active,
binding. The NAME of the tail is 'major-mode, and COMMAND is
the default definition. The MODE-LIST contain always at least the
'major-mode entry.

Initially the alist is nil."
  
  ;Try to make keymap buffer-local, *not* mode-local
  (if (current-local-map)
      (setq minor-local-keymap (copy-keymap (current-local-map)))
    (setq minor-local-keymap (make-sparse-keymap)))
  (use-local-map minor-local-keymap))


(defun minor-set-key (key command name)
  "Bind KEY to COMMAND in the minor keymap, used by the minor mode NAME.
NAME is a symbol that must identify the minor mode."

  (rplaca (create-entry-minor-keymap key)
	  (cons name command))
  (local-set-key key command))

(defun create-entry-minor-keymap (key)
  "Allocate space for a new definition of KEY in the minor mode map.
Return the new entry."

  (if (null minor-local-keymap)
      (make-minor-keymap))
  (let ((entry (assoc key minor-keymap)))
    (if entry
	  (rplacd entry (cons nil (cdr entry)))	;insert empty cons at head
      (setq minor-keymap
	    (cons (cons key
			(list nil (create-major-entry key)))
		  minor-keymap))
      (cdr (car minor-keymap)))))

(defun create-major-entry (key)
  "Return a conscell ('major-mode . COMMAND), where definition is
the command currently bound to the suplied argument KEY."

  (cons 'major-mode (local-key-binding key)))

(defun unbind-minor-mode (name)
  "Remove all references to NAME from the minor keymap."

  (setq minor-keymap 
	(mapcar 'unbind-minor-entry minor-keymap)))

(defun unbind-minor-entry (entry)
  "Remove any references to a minor mode from ENTRY. ENTRY have the
form (KEY . MODE-LIST) where MODE-LIST is a list of (NAME . COMMAND).
NAME is compared with the value of the variable \"name\", and the
conscell is removed if equal. Also if the first entry in MODE-LIST is
removed, KEY is rebound to the COMMAND of the next cell."

  (let ((current (cdr entry))
	(previous entry))
    (if (eq name (car (car current)))
	(local-set-key (car entry) (cdr (car (cdr current)))))
    (while current
      (if (eq name (car (car current)))
	    (rplacd previous (cdr current))
	(setq previous current))
      (setq current (cdr current)))))

