; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         prefix.el
; RCS:          $Header: prefix.el,v 1.2 88/12/19 17:24:07 jbm Exp $
; Locker:       $Locker: jbm $
; Description:  Command to insert a prefix before each line in region.
; File Author:  Jim Moy, CICD-DMS
; Created:      Wed Nov 30 16:34:37 1988
; Modified:     Sat Nov 11 16:34:40 1989 (Darryl Okahata)
; Language:     Emacs-Lisp
; Package:      N/A
;
; (c) Copyright 1988, 1989, 1990, 1991, Hewlett-Packard Company, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Date: Wed, 26 Jul 1989 15:03:53 GMT
;; Subject: Re: help: adding ">" to every line
;; 
;; > Ok all you GE hackers, here's one for you...
;; > 
;; > I'd like to know of an easy way to insert '>' at the beginning of each
;; > line (i.e. when I reply in notes).  In vi it was:  s/^/>/g    but I
;; > don't know how to do that here.
;; > 
;; > Can anyone help?
;; 
;; Well... you could set your region and then do "M-| sed -e 's/^/>/g'"
;; if you wanted to (i.e., run shell-command-on-region).
;; 
;; But here's something a little bit slicker.  I posted this some time
;; ago... I'm kind of surprised it didn't make it into the HUGE
;; distribution given how well Darryl has been keeping up on things.
;; 
;; This will prefix lines in any region you set.  It'll ask you for what
;; you want to prefix the lines with (defaults to "> " if the response is
;; nil), so it's also helpful for commenting out lines in scripts, etc.
;; Take note of the key binding I set at the end of the source, in case
;; you want to bind it elsewhere.
;; 
;; For you afficionados out there, yes, it'd be a little bit easier to
;; use replace-regexp within the save-excursion, but this was a little
;; exercise when I was just learning elisp...
;; 
;; -jbm
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar prefix-lines-string "> "
  "Default string which is inserted by prefix-lines-region.")

(defun prefix-lines-region (start end prefix)
  "Inserts a string at the beginning of each line which has characters
in region.  The string is prompted for and if none is given the value
of prefix-lines-string is used, which defaults to '> '."
  (interactive "r\nsPrefix: ")
  (let* ((str (cond ((zerop (length prefix)) prefix-lines-string)
		    (t prefix)))
	 (strlen (length str)))
    (save-excursion
      (goto-char start)
      (while (< (point) end)
	(beginning-of-line)
	(insert str)
	(forward-line 1)
	(setq end (+ end strlen))))))

(global-set-key "\C-cp" 'prefix-lines-region)
