; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         proto.el
; RCS:          $Header: $
; Description:  Functions to automatically generate and insert ANSI C
;		Prototypes.
; Author:       Darryl Okahata
; Created:      Tue Oct 24 09:40:31 1989
; Modified:     Fri Jan 26 20:12:13 1990 (Darryl Okahata)
; Language:     Emacs-Lisp
; Package:      N/A
;
; (C) Copyright 1989, 1990, 1991, Hewlett-Packard, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; ***** NOTE: The use of these functions requires a program, such as
;;;		mkptypes(1L), to actually generate the ANSI C prototypes.


(defvar proto-tempfile-template "/usr/tmp/proto"
  "The template to use when making a temporary file name for use with
`mkproto'.  This is a prefix passed to make-temp-name."
)

(defvar proto-program "mkptypes"
  "The program to use to generate ANSI C templates."
)

(defvar proto-options (list "-s")
  "This is a list of options to pass to the `proto-program'."
)

(defvar proto-tmp-buffer "*ANSI C prototypes*"
  "The name of a temporary scratch buffer."
)


(defun insert-prototypes ()
  "Generate and insert ANSI C prototype definitions for the current buffer.
The program specified by `proto-program' is used to do the actual prototype
generation.  The options specified by `proto-options' are also passed to the
`proto-program'."
  (interactive "")
  (let ( exec-template
	 prototypes
       )
    (save-excursion
      (save-window-excursion
	(setq tmpbuffer (get-buffer-create proto-tmp-buffer))
	(setq exec-template
	      '(call-process-region (point-min) (point-max) proto-program
				    nil tmpbuffer nil))
	(setq exec-template (append exec-template proto-options))
	(eval exec-template)
	(set-buffer tmpbuffer)
	(goto-char (point-min))
	(replace-regexp "[ \t]*([ \t]*" "(")
	(goto-char (point-min))
	(replace-regexp "[ \t]*)" ")")
	(goto-char (point-min))
	(replace-regexp "[ \t]*," ",")
	(goto-char (point-min))
	(replace-regexp "[ \t]\\([*0-9A-Za-z_]+\\)(" "\t\\1(")
	(setq prototypes (buffer-substring (point-min) (point-max)))
	(kill-buffer tmpbuffer)
      )
    )
    (insert prototypes)
  )
)
