; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         softkey.el
; RCS:          $Header: sfk-def.el,v 2.1 89/04/01 20:57:50 darrylo Exp $
; Description:  A softkey tree structure
; Author:       Rick Kunin, SPE
; Created:      Tue Jan 31 12:50:22 1989
; Modified:     Fri Feb 22 21:27:24 1991 (Darryl Okahata)
; Language:     Emacs-Lisp
; Package:      N/A
;
; (C) Copyright 1989, 1990, 1991, Hewlett-Packard, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; GNU Emacs and this file "softkey.el", is distributed in the hope
;; that it will be useful, but WITHOUT ANY WARRANTY.  No author or
;; distributor accepts responsibility to anyone for the consequences
;; of using it or for whether it serves any particular purpose or
;; works at all, unless he says so in writing.  Refer to the GNU Emacs
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute GNU
;; Emacs and softkey.el, but only under the conditions described in
;; the GNU Emacs General Public License.  A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you can
;; know your rights and responsibilities.  It should be in a file
;; named COPYING.  Among other things, the copyright notice and this
;; notice must be preserved on all copies.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  implements the softkey tree structure loosely based on the one found in
;  editpc. These provide a new user with an easy set of softkeys to
;  work through the normal word processing/text editing functions.  It
;  is based on the load-function-key function provided in some version
;  of gnu-emacs that I picked up from MIT.  
;  The idea of using \033xfunction was taken from Charles Brown.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'sfk-def)

(defvar current-softkeys 'set-function-keys-to-default
  "Function that restores softkeys to their proper state.")

(defun set-function-keys-to-default ()
  "Download the top level function keys to the terminal." 
  (interactive)
  (load-softkey 1 "  Misc" 'set-function-keys-to-misc)
  (load-softkey 2 "Windows" 'set-function-keys-to-window)
  (load-softkey 3 "  File" 'set-function-keys-to-file)
  (load-softkey 4 "Buffers & Blocks" 'set-function-keys-to-buffers)
  (load-softkey 5 " Search" 'set-function-keys-to-search)
  (load-softkey 6 "Movement" 'set-function-keys-to-move1)
  (load-softkey 7 "  Help" 'set-function-keys-to-help)
  (load-softkey 8 "  Exit" 'set-function-keys-to-exit)
  (setq current-softkeys 'set-function-keys-to-default)
  (softkeys-on))

(defun set-function-keys-to-exit ()
  ""
  (interactive "")
  (load-softkey 1 " Really   Exit" 'save-buffers-kill-emacs)
  (load-softkey 2 " " 'keyboard-quit)
  (load-softkey 3 " " 'keyboard-quit)
  (load-softkey 4 " " 'keyboard-quit)
  (load-softkey 5 " " 'keyboard-quit)
  (load-softkey 6 " " 'keyboard-quit)
  (load-softkey 7 " " 'keyboard-quit)
  (load-softkey 8 " Do not   Exit" 'set-function-keys-to-default)
  (setq current-softkeys 'set-function-keys-to-exit)
  (softkeys-on)
)

(defun set-function-keys-to-misc ()
  "Miscellaneous Softkeys menu"
  (interactive)
  (load-softkey 1 " emacs  command" 'execute-extended-command)
  (if (eql major-mode 'text-mode) 
      (load-softkey 2 " format " 'set-function-keys-to-format)
    (load-softkey 2 " " 'keyboard-quit))
  (load-softkey 3 "  UNIX   access" 'set-function-keys-to-hpux)
  (load-softkey 4 " games" 'set-function-keys-to-games)
;;  (load-softkey 5 "Read Mail" "\033xrmail\n")  ;; commented out since prod-eng 
                                                 ;; doesn't use rmail
  (load-softkey 5 " modes " 'set-function-keys-to-modes)
;;(load-softkey 6 "Send Mail" "\033xmail\n")   ;; commented out since prod-eng
                                               ;; uses elm
  (load-softkey 6 " ispell" 'set-function-keys-to-ispell)
  (load-softkey 7 "  Undo" 'undo "\^_")
  (load-softkey 8 "Previous" 'set-function-keys-to-default)
  (setq current-softkeys 'set-function-keys-to-misc)
  (softkeys-on))


(defun set-function-keys-to-hpux ()
  "Operating system access from emacs"
  (interactive)
  (load-softkey 1 "  UNIX  command" 'shell-command)
  (load-softkey 2 "  UNIX   shell" 'shell)
  (load-softkey 3 " Compile" 'compile)
  (load-softkey 4 "directry editor" 'dired)
  (load-softkey 5 "  UNIX   manual" 'manual-entry)
  (load-softkey 6 " print   buffer" 'print-buffer)
  (load-softkey 7 "  Main" 'set-function-keys-to-default)
  (load-softkey 8 "Previous" 'set-function-keys-to-misc)
  (setq current-softkeys 'set-function-keys-to-hpux)
  (softkeys-on))

(defun justify-paragraph ()
  "Just a little something to call fill-paragraph with 
   a prefix to cause justification."
  (interactive)
  (fill-paragraph 0))

(defun set-function-keys-to-format ()
  "Text formattting softkeys for beginners.  
   This should only be called when in text mode"
  (interactive)
  (load-softkey 1 "set left margin" 'softkey-set-left-margin)
  (load-softkey 2 "st right margin"  'softkey-set-right-margin) 
  (load-softkey 3 " center   line" 'center-line)
  (load-softkey 4 "fill ind region" 'fill-individual-paragraphs)
  (load-softkey 5 "  fill  paragrph" 'fill-paragraph)
  (load-softkey 6 " justifyparagrph" 'justify-paragraph)
  (load-softkey 7 "  Main" 'set-function-keys-to-default)
  (load-softkey 8 "Previous" 'set-function-keys-to-misc)
  (setq current-softkeys 'set-function-keys-to-format)
  (softkeys-on))

(defun softkey-set-left-margin ()
  "Helper function for softkeys to set left margin"
  (interactive)
  (let ((default-column (current-column)))
   (setq left-margin (eval-minibuffer "Left margin? " (int-to-string default-column)))))

(defun softkey-set-right-margin ()
  "Helper function for softkeys to set left margin"
  (interactive)
  (let ((default-column (current-column)))
   (setq fill-column (eval-minibuffer "Right margin? " (int-to-string default-column)))))

(defun set-function-keys-to-modes ()
  "Enable various major and minor modes"
  (interactive)
  (load-softkey 1 "  Text    Mode" 'text-mode)
  (load-softkey 2 "    C     Mode" 'c-mode)
  (load-softkey 3 " Picture  Mode" 'picture-mode)
  (load-softkey 4 "Outline   Mode" 'outline-mode)
  (load-softkey 5 " toggle autofill" 'auto-fill-mode)
  (load-softkey 6 " toggle  abbrev" 'abbrev-mode)
  (load-softkey 7 "  Main" 'set-function-keys-to-default)
  (load-softkey 8 "Previous" 'set-function-keys-to-misc)
  (setq current-softkeys 'set-function-keys-to-modes)
  (softkeys-on))


(defun set-function-keys-to-games ()
  "Window sofkeys for beginners"
  (interactive)
  (load-softkey 1 " hanoi" 'hanoi)
  (load-softkey 2 " doctor" 'doctor)
  (load-softkey 3 "  yow" 'yow)
  (load-softkey 4 " gomoku" 'softkey-gomoku)
  (load-softkey 5 "pinhead" 'psychoanalyze-pinhead)
  (load-softkey 6 "tinymud" 'tinymud)
  (load-softkey 7 "  MAIN" 'set-function-keys-to-default)
  (load-softkey 8 "Previous" 'set-function-keys-to-misc)
  (setq current-softkeys 'set-function-keys-to-games)
  (softkeys-on))

(defun set-function-keys-to-ispell ()
  "Window sofkeys for beginners"
  (interactive)
  (load-softkey 1 " ispell  buffer" 'ispell-buffer)
  (load-softkey 2 " ispell  region" 'ispell-region)
  (load-softkey 3 " ispell   word" 'ispell-word)
  (load-softkey 4 " " 'keyboard-quit)
  (load-softkey 5 " " 'keyboard-quit)
  (load-softkey 6 " " 'keyboard-quit)
  (load-softkey 7 "  MAIN" 'set-function-keys-to-default)
  (load-softkey 8 "Previous" 'set-function-keys-to-misc)
  (setq current-softkeys 'set-function-keys-to-ispell)
  (softkeys-on))

(defun softkey-gomoku()
  "Ensure that gomoku gets loaded when called from softkeys"
  (interactive)
  (autoload 'gomoku "gomoku")
  (gomoku))


(defun set-function-keys-to-window ()
  "Window sofkeys for beginners"
  (interactive)
  (load-softkey 1 "  next   window" 'other-window)
  (load-softkey 2 " shrink  window" 'shrink-window)
  (load-softkey 3 "enlarge  window" 'enlarge-window)
  (load-softkey 4 "  split  window" 'split-window-vertically)
  (load-softkey 5 " split   horiz" 'split-window-horizontally)
  (load-softkey 6 " delete  window" 'delete-window)
  (load-softkey 7 " delete  others" 'delete-other-windows)
  (load-softkey 8 "Previous" 'set-function-keys-to-default)
  (setq current-softkeys 'set-function-keys-to-window)
  (softkeys-on))


(defun set-function-keys-to-file ()
  "File sofkeys for beginners"
  (interactive)
  (load-softkey 1 "  edit    file" 'find-file)
  (load-softkey 2 " insert   file" 'insert-file)
  (load-softkey 3 "  read    file" 'find-file-read-only)
  (load-softkey 4 "  save   buffer" 'save-buffer)
  (load-softkey 5 "save w/ new name" 'write-file)
  (load-softkey 6 " write  modified" 'save-some-buffers)
  (load-softkey 7 "  kill   buffer" 'kill-buffer)
  (load-softkey 8 "Previous" 'set-function-keys-to-default)
  (setq current-softkeys 'set-function-keys-to-file)
  (softkeys-on))

(defun set-function-keys-to-buffers ()
  "Buffer softkeys for beginners"
  (interactive)
  (load-softkey 1 "set mark" 'set-mark-command)
  (load-softkey 2 "swap dot & mark" 'exchange-point-and-mark)
  (load-softkey 3 "   cut   region" 'kill-region)
  (load-softkey 4 " paste" 'yank)
  (load-softkey 5 "  list   buffers" 'list-buffers)
  (load-softkey 6 "  goto   buffer" 'switch-to-buffer)
  (load-softkey 7 "  kill   buffer" 'kill-buffer)
  (load-softkey 8 "Previous" 'set-function-keys-to-default)
  (setq current-softkeys 'set-function-keys-to-buffers)
  (softkeys-on))

(defvar softkey-search-regexp nil
  "If not nil then the softkeys will use the regexp searches" )

(defvar softkey-search-incremental t
  "If not nil then the softkeys will do incremental searches\n")

(defun toggle-softkey-search-incremental ()
  "Kludge to redefine the search softkeys when user wants to toggle 
   incremental mode"
  (interactive)
  (setq softkey-search-incremental (not softkey-search-incremental))
  (set-function-keys-to-search))

(defun toggle-softkey-search-regexp ()
  "Kludge to redefine the search softkeys when user wants to toggle 
   incremental mode"
  (interactive)
  (setq softkey-search-regexp (not softkey-search-regexp))
  (set-function-keys-to-search))


(defun set-function-keys-to-search ()
  "Search softkeys for beginners"
  (interactive)
  (cond
   ((and softkey-search-regexp softkey-search-incremental)
    (load-softkey 1 "isearch reg forw" 'isearch-forward-regexp "\033\^s")
    (load-softkey 2 "isearch reg rev" 'isearch-backward-regexp)
    (load-softkey 3 " query  replace " 'query-replace-regexp)
    (load-softkey 4 "replace" 'replace-regexp)
    (load-softkey 6 "Reg Exp    ON" 'toggle-softkey-search-regexp)
    (load-softkey 7 "  Incr     ON" 'toggle-softkey-search-incremental)
    )
   (softkey-search-regexp
    (load-softkey 1 " search reg forw" 're-search-forward)
    (load-softkey 2 " search reg rev" 're-search-backward)
    (load-softkey 3 " query   replace" 'query-replace-regexp)
    (load-softkey 4 "replace" 'replace-regexp)
    (load-softkey 6 "Reg Exp    ON" 'toggle-softkey-search-regexp)
    (load-softkey 7 "  Incr    OFF" 'toggle-softkey-search-incremental)
    )
   (softkey-search-incremental
    (load-softkey 1 "isearch forward" 'isearch-forward "\^s")
    (load-softkey 2 "isearch reverse" 'isearch-backward "\^r")
    (load-softkey 3 " query   replace" 'query-replace)
    (load-softkey 4 "replace" 'replace-string)
    (load-softkey 6 "Reg Exp   OFF" 'toggle-softkey-search-regexp)
    (load-softkey 7 "  Incr     ON" 'toggle-softkey-search-incremental)
    )
   (t 
    (load-softkey 1 " search forward" 'search-forward)
    (load-softkey 2 " search reverse" 'search-backward)
    (load-softkey 3 " query   replace" 'query-replace)
    (load-softkey 4 "replace" 'replace-string)
    (load-softkey 6 "Reg Exp   OFF" 'toggle-softkey-search-regexp)
    (load-softkey 7 "  Incr    OFF" 'toggle-softkey-search-incremental)
    ))
  (load-softkey 5 " " 'keyboard-quit)
  (load-softkey 8 "Previous" 'set-function-keys-to-default) 
  (setq current-softkeys 'set-function-keys-to-search)
  (softkeys-on))



(defun set-function-keys-to-move1 ()
  "Cursor Movement sofkeys for beginners"
  (interactive)
  (load-softkey 1 "forward   word" 'forward-word)
  (load-softkey 2 "backward  word" 'backward-word)
  (load-softkey 3 "forward sentence" 'forward-sentence)
  (load-softkey 4 "backwardsentence" 'backward-sentence)
  (load-softkey 5 "  <--" 'beginning-of-line)
  (load-softkey 6 "  -->" 'end-of-line)
  (load-softkey 7 "  more" 'set-function-keys-to-move2)
  (load-softkey 8 "Previous" 'set-function-keys-to-default)
  (setq current-softkeys 'set-function-keys-to-move1)
  (softkeys-on))

(defun set-function-keys-to-move2 ()
  "Cursor Movement sofkeys for beginners"
  (interactive)
  (load-softkey 1 "forward paragraph" 'forward-paragraph)
  (load-softkey 2 "backwardparagraph" 'backward-paragraph)
;;  (load-softkey 3 "forward   sexp" "\033\^f")
;;  (load-softkey 4 "backward  sexp" "\033\^b")
  (load-softkey 3 "del next  word" 'kill-word)
  (load-softkey 4 "del prev  word" 'backward-kill-word)
  (load-softkey 5 "  goto   line #" 'goto-line)
  (load-softkey 6 "  show  line #'s" 'softkey-show-line-number)
  (load-softkey 7 "  MAIN" 'set-function-keys-to-default)
  (load-softkey 8 "Previous" 'set-function-keys-to-move1)
  (setq current-softkeys 'set-function-keys-to-move2)
  (softkeys-on))


(defun softkey-show-line-number ()
  "Ensure that line-number gets loaded when called from softkeys"
  (interactive)
  (autoload 'display-line-numbers "line-num")
  (display-line-numbers)
  (count-lines-page))


(defun set-function-keys-to-help ()
  "Help sofkeys for beginners"
  (interactive)
  (load-softkey 1 "  list  bindings" 'describe-bindings)
  (load-softkey 2 "describecommand " 'describe-function)
  (load-softkey 3 "describe   key  " 'describe-key)
  (load-softkey 4 "describevariable" 'describe-variable)
  (load-softkey 5 "apropos" 'apropos)
  (load-softkey 6 " where   mapped" 'where-is)
  (load-softkey 7 "  more" 'set-function-keys-to-help2)
  (load-softkey 8 "Previous" 'set-function-keys-to-default)
  (setq current-softkeys 'set-function-keys-to-help)
  (softkeys-on))
  

(defun set-function-keys-to-help2 ()
  "Help sofkeys for beginners"
  (interactive)
  (load-softkey 1 " Manual" 'info)
  (load-softkey 2 "Tutorial" 'help-with-tutorial)
  (load-softkey 3 "Copying"  'describe-copying)
  (load-softkey 4 " " 'keyboard-quit)
  (load-softkey 5 " " 'keyboard-quit)
  (load-softkey 6 " " 'keyboard-quit)
  (load-softkey 7 "  MAIN" 'set-function-keys-to-default)
  (load-softkey 8 "Previous" 'set-function-keys-to-help)
  (setq current-softkeys 'set-function-keys-to-help2)
  (softkeys-on))

(set-function-keys-to-default)
