; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         tm-sfk-fns.el
; RCS:          $Header: tm-sfk-fns.e,v 2.0 89/03/30 12:26:44 darrylo Exp $
; Description:  Functions used to implement softkeys from an HP terminal
; Author:       Rick Kunin, SPE
; Created:      Thu Mar 16 19:11:51 1989
; Modified:     Fri Feb 22 21:29:21 1991 (Darryl Okahata)
; Language:     Emacs-Lisp
; Package:      N/A
;
; (C) Copyright 1989, 1990, 1991, Hewlett-Packard Company, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun function-keys-on ()
  "Turn the function keys on"
  (interactive)
  (send-string-to-terminal "\e&jB"))

(defun function-keys-off ()
  "Turn the function keys off"
  (interactive)
  (send-string-to-terminal "\e&j@"))

(autoload 'set-function-keys-to-default "sfk-def")
(defvar current-softkeys 'set-function-keys-to-default
  "Function that restores softkeys to their proper state.")

(defun tm-toggle-softkeys ()
  "Toggle the softkeys display on a terminal display."
  (interactive)
  (if softkeys-arent-displayed
      (progn
	(softkeys-on)
	(funcall current-softkeys))
    (softkeys-off)))

(defun load-function-key (function-number label string)
  (if (or (< function-number 1) (> function-number 8))
      (error "function key number not in 1-8 range")
    ; terminals don't like labels longer than 16 characters
    (if (> (length label) 16) (setq label (substring label 1 16)))
    (send-string-to-terminal
     (concat "\e&f"			; Escape sequence initializer
	     function-number "k"
	     (length label)  "d"	; Send length in bytes of label string
	     (length string) "L"	; length of defined function sequence
	     label
	     string))))

(defun function-keys-init ()
  "Setup the function key unique key map definitions"
  (interactive)
  (load-function-key 1 " " default-softkey-1)
  (load-function-key 2 " " default-softkey-2)
  (load-function-key 3 " " default-softkey-3)
  (load-function-key 4 " " default-softkey-4)
  (load-function-key 5 " " default-softkey-5)
  (load-function-key 6 " " default-softkey-6)
  (load-function-key 7 " " default-softkey-7)
  (load-function-key 8 " " default-softkey-8)
  )
  

(defun new-load-function-key (function-number label function-name &optional string)
  (if (or (< function-number 1) (> function-number 8))
      (error "function key number not in 1-8 range")
    (progn
      ;; if a string is passed then the softkey is redifined to be
      ;; that string.  If no string is passed then the softkey is
      ;; redefined to the 'default' string and then that default is
      ;; bound to the function-name that is passed in.  If a string is
      ;; passed in then the function name is ignored.
      (if (not string)
	  (progn
	    (cond
	     ((eq function-number 1)
	      (setq string default-softkey-1))
	     ((eq function-number 2)
	      (setq string default-softkey-2 ))
	     ((eq function-number 3)
	      (setq string default-softkey-3 ))
	     ((eq function-number 4)
	      (setq string default-softkey-4 ))
	     ((eq function-number 5)
	      (setq string default-softkey-5 ))
	     ((eq function-number 6)
	      (setq string default-softkey-6 ))
	     ((eq function-number 7)
	      (setq string default-softkey-7 ))
	     ((eq function-number 8)
	      (setq string default-softkey-8 )))
	    (global-set-key string function-name)))
      ; terminals don't like labels longer than 16 characters
      (if (> (length label) 16) (setq label (substring label 1 16)))
      (send-string-to-terminal
       (concat "\e&f"			; Escape sequence initializer
	       function-number "k"
	       (length label)  "d"	; Send length in bytes of label string
	       (length string) "L"      ; length of defined function sequence
	       label
	       string
	     ))
	)))

