;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Copyright (c) 1993, 1994 by William M. Perry (wmperry@indiana.edu)
;;;
;;; This file is part of GNU Emacs.
;;;
;;; GNU Emacs is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;;
;;; GNU Emacs is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to
;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Enhancements For Emacs 19
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; File-name-handler-alist functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(if (w3-rassoc 'w3-file-handler file-name-handler-alist) nil
  (setq file-name-handler-alist
 	(let ((new-handler (cons (concat "^/"
					 (substring w3-nonrelative-link 1 nil))
 				 'w3-file-handler)))
 	  (if file-name-handler-alist
 	      (append (list new-handler) file-name-handler-alist)
 	    (list new-handler)))))

(defun w3-file-handler (operation &rest args)
  "Function called from the file-name-handler-alist routines.  OPERATION
is what needs to be done ('file-exists-p, etc).  args are the arguments
that would have been passed to OPERATION."
  (let ((fn (get operation 'w3-mode))
	(url (substring (car args) 1 nil))
	(myargs (cdr args)))
    (if (string-match (concat w3-nonrelative-link "/[^/]") url)
	(setq url (concat (w3-match url 1) "://"
			  (substring url (1- (match-end 0)) nil))))
    (if fn (apply fn url myargs)
      (let (file-name-handler-alist)
	(apply operation url myargs)))))

(put 'insert-file-contents 'w3-mode 'w3-insert-file-contents)
(put 'directory-files 'w3-mode 'w3-directory-files)
(put 'file-directory-p 'w3-mode 'w3-file-directory-p)
(put 'file-writable-p 'w3-mode 'w3-file-writable-p)
(put 'file-readable-p 'w3-mode 'w3-file-exists)
(put 'file-executable-p 'w3-mode 'null)
(put 'file-symlink-p 'w3-mode 'null)
(put 'file-exists-p 'w3-mode 'w3-file-exists)
(put 'copy-file 'w3-mode 'w3-copy-file)
(put 'file-attributes 'w3-mode 'w3-file-attributes)
(put 'file-name-all-completions 'w3-mode 'w3-file-name-all-completions)
(put 'file-name-completion 'w3-mode 'w3-file-name-completion)
(put 'file-local-copy 'w3-mode 'w3-file-local-copy)
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Menubars, etc.  Cheat and use the lucid compatibility package, because
;;; I find it much more readable then the emacs19 keymap defs. ugh.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar w3-links-menu nil "Menu for w3-mode in emacs 19")
(make-variable-buffer-local 'w3-links-menu)

(defvar w3-options-menu
  '("Options"
    ["Delay Image Load" (setq w3-delay-image-loads (not w3-delay-image-loads))
     nil]
    ["Flush Image Cache" (setq w3-graphics-list nil) t]
    ("Hypertext Gopher Mode"
     ["Turn On" (setq w3-use-hypertext-gopher t) t]
     ["Turn Off" (setq w3-use-hypertext-gopher nil) t])
    ("Hypertext Dired Mode"
     ["Turn On" (setq w3-directory-format 'hypertext) t]
     ["Turn Off" (setq w3-directory-format 'dired) t])
    ["Clear History" (setq w3-history-list nil) t])
  "The options menu for w3")

(defun w3-create-faces ()
  "Create faces, the emacs 19 way"
  (if (or (eq window-system 'x)		; X-windows?
	  (eq window-system 'ns)	; NeXTstep?
	  (eq window-system 'pm)	; OS/2 Presentation manager?
	  )
      (progn
	(make-face w3-default-style)
	(make-face w3-header-style)
	(make-face w3-bold-style)
	(make-face w3-italic-style)
	(make-face w3-underline-style)
	(make-face w3-node-style)
	(make-face w3-address-style)
	(make-face w3-tt-style)
	(make-face w3-visited-node-style)
	(if (not (face-differs-from-default-p w3-header-style))
	    (copy-face 'bold-italic w3-header-style))
	(if (not (face-differs-from-default-p w3-visited-node-style))
	    (copy-face 'bold-italic w3-visited-node-style))
	(if (not (face-differs-from-default-p w3-node-style))
	    (copy-face 'bold w3-node-style))
	(if (not (face-differs-from-default-p w3-address-style))
	    (copy-face 'italic w3-address-style))
	(if (not (face-differs-from-default-p w3-bold-style))
	    (copy-face 'bold w3-bold-style))
	(if (not (face-differs-from-default-p w3-italic-style))
	    (copy-face 'italic w3-italic-style))
	(if (not (face-differs-from-default-p w3-tt-style))
	    (copy-face 'default w3-tt-style))
	(if (not (face-differs-from-default-p w3-underline-style))
	    (copy-face 'underline w3-underline-style)))))

(defun w3-zone-start (zone)
  "This is just a hack to get the beginning of a text-property region"
  (let ((link (get-text-property zone 'w3))
	(form (get-text-property zone 'w3form)))
    (cond
     (link (or (previous-single-property-change zone 'w3) (point-min)))
     (form (or (previous-single-property-change zone 'w3form) (point-min)))
     (t (error "No zone contains %d!" zone)))))

(defun w3-zone-end (zone)
  "This is just a hack to get the end of a text-property region"
  (let ((link (get-text-property zone 'w3))
	(form (get-text-property zone 'w3form)))
    (cond
     (link (or (next-single-property-change zone 'w3) (point-max)))
     (form (or (next-single-property-change zone 'w3form) (point-max)))
     (t (error "No zone contains %d!" zone)))))

(defun w3-zone-eq (zone1 zone2)
  "Are two zones equal?"
  (and
   (= (w3-zone-start zone1) (w3-zone-start zone2))
   (= (w3-zone-end zone1) (w3-zone-end zone2))))

(defun w3-fix-extent-endpoints ()
  "Make sure no extents have whitespace/newlines at the end of them"
  (let ((zones (w3-all-forms-zones)) nd)
    (while zones
      (setq nd (1- (w3-zone-end (car zones))))
      (while (equal (char-after nd) ?\n)
	(set-text-properties nd (1+ nd) nil)
	(setq nd (1- nd)))
      (setq zones (cdr zones)))))
      
(defun w3-delete-zone (zone)
  "Remove a zone from a buffer.  This is just a hack to remove some
text properties."
  (let ((st (w3-zone-start zone))
	(nd (w3-zone-end zone)))
    (remove-text-properties st nd
			    (list 'w3 nil
				  'w3form nil
				  'face nil
				  'mouse-face nil))))

(defun w3-zone-data (zone)
  "Return the data from a zone (fake text property area"
  (let ((link (get-text-property zone 'w3))
	(form (get-text-property zone 'w3form))
	(grph (get-text-property zone 'w3graphic))
	(list (get-text-property zone 'w3expandlist)))
    (cond
     (link (cons 'w3 link))
     (form (cons 'w3form form))
     (grph (cons 'w3graphic grph))
     (list (cons 'w3expandlist list))
     (t nil))))

(defun w3-only-links ()
  "Get all the zones from a buffer.  This is really just a list of positions
that have a text-property set correctly."
  (let ((cur (point-min))
	(result '()))
    (while cur
      (if (get-text-property cur 'w3)
	  (setq result (cons cur result)))
      (setq cur (next-single-property-change cur 'w3)))
    (nreverse result)))

(defun w3-all-forms-zones ()
  "Get all the zones from a buffer.  This is really just a list of positions
that have a text-property set correctly."
  (let ((cur (point-min))
	(result '()))
    (while cur
      (if (get-text-property cur 'w3form)
	  (setq result (cons (1+ cur) result)))
      (setq cur (next-single-property-change cur 'w3form)))
    (nreverse result)))

(defun w3-all-zones ()
  "Get all the zones from a buffer.  This is really just a list of positions
that have a text-property set correctly."
  (nconc (w3-only-links) (w3-all-forms-zones)))
   
(defun w3-add-hotlist-menu ()
  "Add the hotlist menu to this buffer - used when it changes."
  (if (eq major-mode 'w3-mode)
      (let ((hot-menu nil)
	    (hot w3-hotlist))
	(while hot
	  (setq hot-menu 
		(cons (vector (car (car hot))
			      (list 'w3-fetch (car (cdr (car hot))))
			      t) hot-menu)
		hot (cdr hot)))
	(add-menu '("Navigate") "Hotlist..." hot-menu))))

(defun w3-find-specific-link (link)
  "Find LINK in the current document"
  (let ((cur (point-min))
	found
	next-change)
    (while (or (setq next-change (next-single-property-change cur 'w3-ident))
	       (not found))
      (if (equal link (get-text-property cur 'w3-ident))
	  (progn
	    (setq found (goto-char cur)
		  cur (point-max))
	    (message "Found link %s" link))
	(setq cur next-change)))))

(defun w3-link-at (pt)
  "Return the link(s) at point"
  (get-text-property pt 'w3))

(defun w3-forward-form-internal (&optional pt)
  "Go forward 1 link"
  (if (null pt) (setq pt (point)))
  (let ((next (next-single-property-change pt 'w3form)))
    (cond
     ((null next) nil)
     ((get-text-property next 'w3form) next)
     (t (w3-forward-form-internal next)))))

(defun w3-forward-link-internal (&optional pt)
  "Go forward 1 link"
  (if (null pt) (setq pt (point)))
  (let ((next (next-single-property-change pt 'w3)))
    (cond
     ((null next) nil)
     ((get-text-property next 'w3) next)
     (t (w3-forward-link-internal next)))))

(defun w3-forward-link ()
  "Go forward 1 link."
  (interactive)
  (let ((link (w3-forward-link-internal))
	(form (w3-forward-form-internal)))
    (cond
     ((and (null link) (null form)) (error "No more links."))
     (t
      (goto-char (min (or link (point-max)) (or form (point-max))))
      (skip-chars-forward " \t\n")))))

(defun w3-back-link-internal (&optional pt)
  (if (and (null pt) (get-text-property (point) 'w3)
	   (previous-single-property-change (point) 'w3))
      (setq pt (1- (previous-single-property-change (point) 'w3))))
  (if (null pt) (setq pt (point)))
  (if (null (previous-single-property-change pt 'w3)) nil
    (let ((prev (1- (previous-single-property-change pt 'w3))))
      (cond
       ((null prev) nil)
       ((get-text-property prev 'w3)
	(let ((prev2 (previous-single-property-change prev 'w3)))
	  (if prev2 prev2 (point-min))))
       (t (w3-back-link-internal prev))))))

(defun w3-back-form-internal (&optional pt)
  (if (and (null pt) (get-text-property (point) (quote w3form)))
      (setq pt (previous-single-property-change (1- (point)) (quote w3form))))
  (if (null pt) (setq pt (point)))
  (let ((prev (previous-single-property-change pt (quote w3form))))
    (cond
     ((null prev) nil)
     ((get-text-property prev (quote w3form))
      (let ((prev2 (previous-single-property-change prev (quote w3form))))
	(if prev2 prev2 (point-min))))
     (t (w3-back-form-internal (1- prev))))))

(defun w3-back-link ()
  "Go back 1 link"
  (interactive)
  (let ((link (w3-back-link-internal))
	(form (w3-back-form-internal)))
    (if (equal form (point)) (setq form nil))
    (if (equal link (point)) (setq link nil))
    (cond
     ((and (null link) (null form)) (error "No previous links."))
     (t
      (goto-char (max (or link (point-min)) (or form (point-min))))
      (skip-chars-forward " \t\n")))))

(defun w3-zone-hidden-p (start end)
  "Return t iff the region from start to end is invisible."
  (get-text-property (1+ start) 'invisible))

(defun w3-unhide-zone (start end)
  "Make a region from START TO END visible. (emacs19)"
  (add-text-properties start end (list 'invisible nil)))

(defun w3-hide-zone (start end)
  "Make a region from START to END invisible. (emacs19)"
  (add-text-properties start end (list 'invisible t)))

(defun w3-add-zone (start end style data &optional highlight)
  "Add highlighting (emacs19)"
  (add-text-properties start end
		       (list 'face style ; Set the font, etc
			     (car data) (cdr data)))
  (if (and (eq (car data) 'w3) (nth 1 data))
      (add-text-properties start end (list 'w3-ident (nth 1 data))))
  (if (eq (car-safe data) 'w3form)
      (add-text-properties start end (list 'front-sticky nil
					   'rear-nonsticky '(t)))))

(defun w3-follow-mouse (e)
  "Function suitable to being bound to a mouse key.  Follows the link under
the mouse click."
  (interactive "e")
  (mouse-set-point e)
  (w3-follow-link))

(defun w3-zone-at (pt)
  "Return the extent at point PT that is either a link or a forms area."
  (let ((link (get-text-property pt 'w3))
	(form (get-text-property pt 'w3form))
	(grph (get-text-property pt 'w3graphic))
	(list (get-text-property pt 'w3expandlist)))
    (cond
     (link pt)
     (form pt)
     (grph pt)
     (list pt)
     (t nil))))

(defun w3-follow-inlined-image-mouse (e)
  "Follow the inlined image under the mouse - ignore any hyperlinks or
form entry areas and blindly try to find an image."
  (interactive "e")
  (mouse-set-point e)
  (w3-follow-inlined-image))

(defun w3-follow-inlined-image ()
  "Follow an inlined image, regardless of whether it is a hyperlink or not."
  (interactive)
  (let ((grph (get-text-property (point) 'w3graphic)))
    (cond
     (grph (w3-maybe-relative (nth 0 grph)))
     (t (error "No inlined image at point.")))))

(if (boundp 'hyperb:version)
    nil
  (define-key w3-mode-map [mouse-2] 'w3-follow-mouse)
  (define-key w3-mode-map [mouse-3] 'w3-popup-menu))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Functions to build menus of urls
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun w3-build-FSF19-menu ()
  "Build emacs19 menus from w3-links-list"
  (let* ((hot w3-hotlist)
	 (hot-menu nil)
	 (ovls (w3-only-links)))
    (while ovls
      (let ((data (w3-zone-data (car ovls))))
	(if (and (eq (car-safe data) 'w3) (nth 2 data))
	    (setq w3-links-menu 
		  (nconc w3-links-menu
			 (list
			  (vector
			   (format "%s" (w3-strip-leading-spaces (nth 3 data)))
			   (list 'w3-maybe-relative (nth 2 data)) t))))))
      (setq ovls (cdr ovls)))
    (setq w3-links-menu (cons "Links..." w3-links-menu))
    (while hot
      (setq hot-menu
	    (cons (vector (car (car hot))
			  (list 'w3-maybe-relative (car (cdr (car hot))))
			  t) hot-menu))
      (setq hot (cdr hot)))
    (setq hot-menu (cons "Hotlist..." hot-menu))
    (set-buffer-menubar (copy-sequence current-menubar))
					;    (add-menu '("Help") "WWW" (cdr w3-help-menu))
    (add-menu nil "WWW" (cdr w3-popup-menu))
    (add-menu nil "Navigate" (append
			      (cdr w3-navigate-menu)
			      (list hot-menu w3-links-menu)))
    (add-menu nil "Options" (cdr w3-options-menu))
    (add-menu nil "Annotate" (cdr w3-annotation-menu))))

(defun w3-popup-menu (e)
  "Pop up a menu of common w3 commands"
  (interactive "e")
  (mouse-set-point e)
  (popup-menu w3-popup-menu))

(defun w3-insert (&rest args)
  "Insert without inheriting any text properties"
  (let ((start (point)))
    (apply 'insert-before-markers args)
    (set-text-properties start (point) nil)))

(defun w3-setup-version-specifics ()
  "Set up routine for emacs 19"
  (if (memq window-system
	    '(x pm ns))       ;; Only load this up in X, presentation manager,
			      ;; or NeXTstep, otherwise emacs will barf
      (let ((saved-menu-info (mapcar
			      (function
			       (lambda (x)
				 (cons x (cdr (assoc 'menu-bar-lines
						     (frame-parameters x))))))
			      (frame-list))))
	(progn
	  (require 'menubar "lmenu")
	  (require 'lucid))
	(mapcar (function
		 (lambda (x)
		   (modify-frame-parameters (car x) (list (cons 'menu-bar-lines
								(cdr x))))))
		saved-menu-info)))
  (cond 
   ((or (eq window-system 'x) (eq window-system 'pm))
    (fset 'w3-x-popup-menu 'x-popup-menu))
   ((eq window-system 'ns)
    (fset 'w3-x-popup-menu 'ns-popup-menu))))

(defun w3-store-in-x-clipboard (str)
  "Store string STR in the Xwindows clipboard"
  (cond
   ((eq window-system 'x) (x-set-selection 'PRIMARY str))
   ((eq window-system 'pm) (x-set-selection 'PRIMARY str))
   ((eq window-system 'ns) (ns-store-pasteboard-internal str))
   (t nil)))

(defun w3-mode-version-specifics ()
  "Emacs 19 specific stuff for w3-mode"
  (if (or (memq window-system '(x pm ns)))
      (w3-build-FSF19-menu)))

(defun w3-map-links (function &optional buffer from to maparg)
  "Map FUNCTION over the hypertext links which overlap region in BUFFER,
starting at FROM and ending at TO.  FUNCTION is called with the arguments
linkdata, MAPARG.
The arguments FROM, TO, MAPARG, and BUFFER default to the beginning of
BUFFER, the end of BUFFER, nil, and (current-buffer), respectively.

In emacs19, FROM, TO, and BUFFER are ignored.... working on it."
  (mapcar (function (lambda (x)
		      (funcall function (w3-zone-data x) maparg)
		      nil)) (w3-only-links))
  nil)

(provide 'w3-emacs19)
