;;; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; File:         xmh-mode.el
;;; RCS:          $Header: xmh-mode.el,v 1.2 89/09/22 13:11:03 brown PRODUCTION $
;;; Description:  Special mode for composing and sending xmh drafts.  The
;;;               origins of this code were once mh-e.el
;;; Author:       Niels Mayer, HPL
;;; Created:      Fri May  8 18:47:13 1987
;;; Modified:     Sat Feb 13 15:58:59 1993 (Carl Edman) cedman@capitalist
;;; Language:     GNU Emacs Lisp
;;; Package:      GNU Emacs Version 18.54
;;;
;;; (c) Copyright 1987, 1988, 1989, 1990, 1991, Hewlett-Packard Company, all rights reserved.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;

(setq xmhmol '(
  "@(#) CCC-SRC $Header: xmh-mode.el,v 1.2 89/09/22 13:11:03 brown PRODUCTION $"
  ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 890816 cc   Fixes for use with latest gnu emacs (not gnuvo) 18.54.
;;; 890818 cc   Changed "ivomh" to "xmh" globally.
;;; 890821 cc   Finishing touches to buffer/window handling.
;;;             Added Jim Moy's "mh-mail-header-tabs" function and
;;;             bound it to tab, and incorporated his "mh-moon-header"
;;;             into this file.  Avoid err msg in moon-header in case
;;;             separator line "^--------" not found.

;;; Use the following X defaults:
;;; Xmh*whatNowProc:                gnudraft
;;; Xmh*compProc:                   gnudoit "(xmh-smail)"
;;; Xmh*editProc:                   gnuclient
;;; Xmh*compUseProc:                gnuclient

;;; To initialize xmh-mode, you have to call xmh-mode-init instead of
;;; xmh-mode

;;; Set for local environment:
;;;* These are now in paths.el.
;;;(defvar mh-progs "/usr/local/bin/mh/" "Directory containing MH commands")
;;;(defvar mh-lib   "/usr/local/lib/mh/" "Directory of MH library")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Initialization Function and global Variables to initialize.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar xmh-home-dir ""
  "The home directory on the local machine or the machine running xmh"
  )

(defvar xmh-mail-dir ""
  "The mh mail directory on the local machine or the machine running xmh."
  )

(defvar xmh-drafts-dir ""
  "The mh drafts directory on the local machine or the machine running xmh."
  )

(defun xmh-mode-init (&optional homedir)
  "This function must be called in your ~/.emacs in order to initialize some
variables used by the functions in this module. If you run xmh and gnu emacs on 
the same macine, then just include the following form in ~/.emacs:
     (xmh-mode-init)
Users running gnu emacs on a local machine and xmh on a remote machine may prefer
to use the local emacs for editing xmh drafts. In that case, you'll need
to tell xmh-mode the remote machine's home directory so that xmh-mode's 
functions access the mh folders on the remote machine. In this case, in your 
.emacs (or a 'chmod go-r' file that you load from .emacs) you should do 
something like:
     (setq server-netunam-alist '((\"/net/machinez\" . \"login:passwd\"))
     (netunam \"/net/machinez\" \"login:passwd\")
     (xmh-mode-init \"/net/machinez/va/john_doe/\")"
  
  (if homedir
      (setq xmh-home-dir homedir)
    (setq xmh-home-dir (format "%s/" (getenv "HOME")))
    )
  (if (equal (setq xmh-mail-dir (xmh-get-profile-field "Path:")) "")
      (setq xmh-mail-dir "Mail/")
    (setq xmh-mail-dir (format "%s/" xmh-mail-dir))
    )
  (if (not (equal (substring xmh-mail-dir 0 1) "/"))
      (setq xmh-mail-dir (format "%s%s" xmh-home-dir xmh-mail-dir))
    )
  (if (equal (setq xmh-drafts-dir (xmh-get-profile-field "Draft-Folder:")) "")
      (setq xmh-drafts-dir "drafts/")	                                        ;default if nonexistsnt
    (setq xmh-drafts-dir (format "%s/" (substring xmh-drafts-dir 1 nil)))	;else read .mh_profile folder name sans '+'
    )
  (setq xmh-drafts-dir (format "%s%s" xmh-mail-dir xmh-drafts-dir))
  
  (setq xmh-drafts-re (format "%s[0-9]*$" xmh-drafts-dir))
  (setq auto-mode-alist (cons (cons xmh-drafts-re 'xmh-draft-mode)
			      auto-mode-alist))
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Xmh-draft-mode, the mode assumed by all the functions following this.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun xmh-draft-mode ()
  "A Mode for composing and sending mail drafts using MH's 
drafts folder facility. 

The following keybindings are used in this mode:
\\{xmh-draft-mode-map}
xmh-draft-mode-hook is a user-customizable
variable that is funcalled on new xmh-drafts that have just been
brought up. You can use it to do obscure and stupid things like 
automatically inserting a signature at the end of the buffer, or a line 
indicating the phase of the moon in the header.

Example:
    (setq xmh-draft-mode-hook 'xmh-moon-header)
"
  (text-mode)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^[ \t]*[-_][-_][-_]+$\\|" paragraph-start))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate
	(concat "^[ \t]*[-_][-_][-_]+$\\|" paragraph-separate))
  (make-local-variable 'make-backup-files) 
  (setq make-backup-files nil)		;don't make backups of drafts
  (make-local-variable 'file-precious-flag) 
  (setq file-precious-flag t)		;why not
  (make-local-variable 'auto-save-visited-file-name) 
  (setq auto-save-visited-file-name t)	;autosave by overwrite
  (use-local-map xmh-draft-mode-map)
  (setq major-mode 'xmh-draft-mode)
  (setq mode-name "xmh-draft")
  (setq mode-line-buffer-identification'("draft(%b) ^C^S(end) ^C^Q(uit) ^C^K(ill)"))

  ;; the following buffer-local-variables take the place of the environment variables
  ;; passed by the mh whatnow(1) program to the mh send(1) program. See the function
  ;; xmh-exit-sending-draft to see how these are used in passing them to send(1).
  ;; See also the function xmh-create-draft-buffer, called from gnudraft.sh,
  ;; the xmh/gnu emacs equivalent of the mh whatnow(1) program.
  (make-local-variable 'mhfolder)	;path to the altmessage folder -- dist, repl. Used in conjunction w/ mhmessages to give list of msgs to annotate.
  (make-local-variable 'mhaltmsg)	;path to altmessage -- dist, repl. This is how dist gives the message for redistribution to send
  (make-local-variable 'mhdist)		;zero unless "dist" is called
  (make-local-variable 'mhmessages)	;list of messages to annotate. Folder containing these is mhfolder.
  (make-local-variable 'mhannotate)	;annotation type = Resent, Forwarded, Replied.
  (make-local-variable 'mhinplace)	;set if annotation to occur inplace.
  (make-local-variable 'send-watch-p) ;t if ^u^c^s to send mail.
  (run-hooks 'xmh-draft-mode-hook)
  )

(defvar xmh-draft-mode-map (make-sparse-keymap) 
  "Keymap for composing mail."
  )

(defvar xmh-draft-mode-hook ()
  "You may bind this to a function that is invoked upon a new xmh-draft
when the draft buffer is brought up."
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Interface to gnudraft.sh
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun xmh-create-draft-buffer(draft folder altmsg dist messages annotate inplace) 
  "This function is called by the shell script gnudraft, which
acts as a proper mh whatnowproc for gnuemacs. The parameters represent the envars
passed into the mh whatnow program. They are made buffer-local-variables in the
draft buffer and are then passed on as environment variables to mh's send in
function xmh-exit-sending-draft. If any of these variables are nil, then
that environment variable was not present. This function calls 
xmh-draft-buffer-hooks. The default hook is the function 
xmh-create-draft-buffer-default-hook. See it's documentation for more details."
  ;;(save-excursion
    ;; (set-buffer (find-file draft))
    (find-file draft)
    (xmh-draft-mode)
    ;; set the buffer local variables (see fn xmh-draft-mode).
    (setq mhfolder folder)
    (setq mhaltmsg altmsg)
    (setq mhdist dist)
    (setq mhmessages messages)
    (setq mhannotate annotate)
    (setq mhinplace inplace)
    (run-hooks 'xmh-create-draft-buffer-hooks)
    (server-make-window-visible)
   ;;)
  )

(defvar xmh-create-draft-buffer-hooks '(xmh-create-draft-buffer-default-hook)
  "These hooks are called in the function xmh-create-draft-buffer which in
turn is called by the C routine gnudraft.sh. Gnudraft acts as a 
proper whatnowproc for gnu emacs. The hook that is called by default in 
xmh-create-draft-buffer-hooks is xmh-create-draft-buffer-default-hook. See
its documentation for more details."
  )

(defun xmh-create-draft-buffer-default-hook ()
  "This function automatically places the cursor on the correct field of
the draft by looking at the mhannotate buffer-local-variable in an 
xmh-draft-mode buffer. The value of mhannotate decides whether
the draft is generated by repl, forw, comp, or dist. In order for this
to work correctly, you must specify '-annotate' for 'repl' 'forw' 'dist' 
in your ~/.mh_profile."
  (cond
   ((string= mhannotate "Replied")	;generated by "repl -annotate"
    (xmh-goto-header-end 1)		;goto first line after header end
    )
   ((string= mhannotate "Forwarded")	;generated by "forw -annotate"
    (let ((last-input-char ?t)) (xmh-to-field)) ;goto "To: field"
    )
   ((string= mhannotate "Resent")	;generated by "dist -annotate"
    (goto-char (point-min))		;goto end of first line
    (end-of-line)
    (if (not (looking-at "[ \t]")) 
	(insert-string " "))
    )
   (t					;default case, for "comp" and others?
    (let ((last-input-char ?t)) (xmh-to-field)) ;goto "To: field"
    )
   )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun xmh-smail ()
  "A faster version of comp that can be called from xmh by setting 
the X resource \"Xmh*compProc: gnudoit '(xmh-smail)'\". It also allows
a quickie send draft to be brought up directly from within emacs."
  (interactive)
  (let ((i 1)
	(draft-template-file nil)
	(draft-file nil))
    (cond 
     ((file-exists-p (setq draft-template-file (concat xmh-mail-dir "components"))))
     ((file-exists-p (setq draft-template-file (concat mh-lib "components"))))
     (t (message "Can't find file %s/components, check variable mh-lib or install mh."))
     )
    (while (file-exists-p (setq draft-file (concat xmh-drafts-dir i))) (setq i (+ i 1)))
    (copy-file draft-template-file draft-file)
    ;; (set-buffer (find-file draft-file))
    (find-file draft-file)
    (server-make-window-visible)
    (let ((last-input-char ?t)) (xmh-to-field))
    )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun xmh-to-field ()
  "Move point to the end of the header field indicated by the previous
keystroke. Create the field if it does not exist. Set the mark to the
point before moving."
  (interactive "")
  (expand-abbrev)
  (let ((target (cdr (assoc last-input-char
			    '((?t . "To:") (?s . "Subject:") (?c . "Cc:")
			      (?b . "Bcc:") (?f . "Fcc:"))))))
    (cond 
     ((xmh-position-on-field target t)
      (if (not (looking-at "[ \t]")) 
	  (insert-string " "))
      )
     (t
      (goto-char (point-min))
      (re-search-forward "^To:")
      (forward-line 1)
      (insert-string (format "%s \n" target))
      (backward-char 1)
      )
     )
    ))

(defun xmh-position-on-field (field set-mark)
  "Set point to the end of the line beginning with FIELD.
Set the mark to the old value of point, if SET-MARK is non-nil."
  (if set-mark (push-mark))
  (goto-char (point-min))
  (xmh-goto-header-end 0)
  (if (re-search-backward (format "^%s" field) nil t)
      (progn (end-of-line) t)
      nil))

(defun xmh-goto-header-end (arg)
  "Find the end of the message header in the current buffer and position
the cursor at the ARG'th newline after the header."
  (if (re-search-forward "^$\\|^-+$" nil nil)
      (forward-line arg)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun xmh-exit-sending-draft (&optional arg)
  "Send the draft letter in the current buffer. Calling this with an argument
prefix (typically ^U) will leave the draft buffer up and will allow you to 
watch the mail being delivered."
  (interactive "P")
  (set-buffer-modified-p t)		; Make sure buffer is written
  (save-buffer)
  (message "Sending...")
  (goto-char (point-max))
  (cond
   (arg
    (setq send-watch-p t)
    (insert "\n\n********************* MESSAGE DELIVERY STATUS OUTPUT **************************\n")
    )
   (t
    (setq send-watch-p nil)
    (insert "\n\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    (insert   "\n!!! ERROR IN SEND -- please fix the message headers, delete this            !!!")
    (insert   "\n!!!                  error message, and try sending again.                  !!!")
    (insert   "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n")
    )
   )
  (set-process-sentinel 
   (apply 'start-process 
	  (apply 'append
		 '("send")		;program name
		 (list (buffer-name))	;error output goes to top of buffer
		 '("/bin/env")		;assume HPUX, which has this function
		 (if mhfolder
		     (list (format "mhfolder=%s" mhfolder))
		   nil)
		 (if mhaltmsg
		     (list (format "mhaltmsg=%s" mhaltmsg))
		   nil)
		 (if mhdist
		     (list (format "mhdist=%s" mhdist))
		   nil)
		 (if mhmessages
		     (list (format "mhmessages=%s" mhmessages))
		   nil)
		 (if mhannotate
		     (list (format "mhannotate=%s" mhannotate))
		   nil)
		 (if mhannotate
		     (list (format "mhinplace=%s" mhinplace))
		   nil)
		 '("/usr/local/bin/mh/send")
		 '("-format")		;format the outgoing message
		 '("-noforward")	;don't forward: see process-sentinel-send
		 '("-nopush")		;start-process works as if background ps
		 '("-msgid")		;include unique msgid
		 (list (if arg "-verbose" "-noverbose")) ;wank
		 (list (if arg "-watch" "-nowatch")) ;wank, wank
		 (list (buffer-file-name)) ;the name of the draft file to send
		 ()))
   'xmh-process-sentinel-send)
  (if (not arg)
      ;; Hide buffer unless user wants to watch delivery (if arg). The
      ;; process sentinel will kill the buffer or pop up buf w/ error msg 
      ;; depending on the exit status of send.
      ;; (x-hide-buffer (buffer-name))	
      (switch-to-buffer (other-buffer (buffer-name)))
    )
  )

(defun xmh-process-sentinel-send (process change)
  (cond 
   ((and (memq (process-status process) '(exit signal))
	 (equal (process-exit-status process) 0)) ;successful send
    (save-excursion
      (set-buffer (process-buffer process))
      (set-buffer-modified-p nil))
    (if (not send-watch-p)		;delete the draft buffer only if user-
	(kill-buffer (process-buffer process)) ;-did not want to watch the send.
      )
    )
   (t					          ;send exited abnormally
    ;; (find-buffer (process-buffer process))        ;so pop up the send buffer
    (switch-to-buffer (process-buffer process))   ;so pop up the send buffer
    (beep)				          ;with the errmsg.
    )
   ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun xmh-exit-preserving-draft (&optional arg)
  "Save the draft in the mh draftsfolder, and exit by killing the buffer."
  (interactive "P")
  (cond
   ((y-or-n-p "Quit the mail draft? ")
    (set-buffer-modified-p t)		; Make sure buffer is written
    (save-buffer)
    (message "Saved mail draft %s to send later..." (buffer-file-name))
    (kill-buffer (buffer-name))
    )
   (t
    (message "Aborting quit -- you may continue editing the mail draft.")
    )
   )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun xmh-exit-killing-draft (&optional arg)
  "Delete the draftfile, and exit by killing the buffer."
  (interactive "P")
  (cond
   ((y-or-n-p "Kill (delete) the draft? ")
    (set-buffer-modified-p nil)		; allow buffer to be killed.
    (delete-file (buffer-file-name))
    (kill-buffer (buffer-name))
    )
   (t
    (message "Aborting kill -- you may continue editing the mail draft.")
    )
   )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(setq xmh-insert-message-into-draft-folder-name "+inbox")
(defvar xmh-insert-message-into-draft-show-command "show -form mhl.repl"
  "The mh show command run by xmh-insert-message-into-draft.")

(defun xmh-insert-message-into-draft (folder msg)
  "Inserts the desired message into the draft at the point, leaving mark at the
end of the insertion. You may customize the mh show command used by changing
the variable xmh-insert-message-into-draft-show-command."
  (interactive (list (read-string "Folder name: " xmh-insert-message-into-draft-folder-name)
		     (read-string "Message(s): " "")))
  (setq xmh-insert-message-into-draft-folder-name folder)
  (cond
   ((string-match "/net" default-directory)
    (let ((host (substring default-directory 
			  (+ 1 (match-end 0)) 
			  (string-match "/" default-directory (+ 1 (match-end 0))))))
      (let ((default-directory (getenv "HOME"))) ;to prevent shell bug caused by pwd being a /net directory
	(shell-command (concat "/usr/bin/remsh " host " " 
			       xmh-insert-message-into-draft-show-command " " 
			       folder " " msg) t)))
    )
   (t 
    (shell-command (concat xmh-insert-message-into-draft-show-command " "
			   folder " " msg) t)
   ))
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun xmh-check-whom ()
  "List recipients of the current letter."
  (interactive)
  (let ((file-name (buffer-file-name)))
    (set-buffer-modified-p t)		; Force writing of contents
    (save-buffer)
    (message "Checking recipients...")
    (switch-to-buffer-other-window "*Mail Recipients*")
    (bury-buffer (current-buffer))
    (delete-region (point-min) (point-max))
    (xmh-exec-cmd-output "whom" t file-name) ;!!!???!!!
    (other-window -1)
    (message "Checking recipients...done")))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; MISCELLANEOUS FUNCTIONS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xmh-get-profile-field (field)
  "Return FIELD from the user's .mh_profile file."
  (save-window-excursion
    (if (not (file-exists-p (format "%s.mh_profile" xmh-home-dir)))
	(error "xmh-mode.el: Cannot find %s.mh_profile. You must run /usr/local/lib/mh/install-mh before using mh and/or Xmh!" xmh-home-dir))
    (switch-to-buffer " *mh_temp*")	; Create if necessary
    (delete-region (point-min) (point-max))
    (insert-file-contents (expand-file-name (format "%s.mh_profile" xmh-home-dir)))
    (xmh-get-field field)
    )
  )

(defun xmh-get-field (field)
  "Find and return the value of field FIELD in the current buffer.
Returns the empty string if the field is not in the message."
  (let ((case-fold-search t))
    (goto-char (point-min))
    (cond ((not (search-forward field nil t)) "")
	  ((looking-at "[\t ]*$") "")
	  (t
	   (re-search-forward "[\t ]*\\([^\t \n].*\\)$" nil t)
	   (let ((field (buffer-substring (match-beginning 1) (match-end 1)))
		 (end-of-match (point)))
	     (forward-line)
	     (while (looking-at "[ \t]") (forward-line 1))
	     (backward-char 1)
	     (format "%s%s" field (buffer-substring end-of-match (point)))
	     )
	   ))
    )
  )

(defun xmh-exec-cmd-output (command display &rest args)
  "Execute MH command COMMAND with DISPLAY flag and ARGS putting the output
into buffer after point.  Set mark after inserted text."
  (push-mark (point) t)
  (apply 'call-process
	 (nconc (list (format "%s%s" mh-progs command) nil t display)
		(xmh-list-to-string args)))
  (exchange-point-and-mark))

(defmacro push (v l)
  (list 'setq l (list 'cons v l)))

(defun xmh-list-to-string (l)
  "Flattens the list L and makes every element of the new list into a string."
  (let ((new-list nil))
    (while l
      (cond ((null (car l)))
	    ((symbolp (car l)) (push (format "%s" (car l)) new-list))
	    ((numberp (car l)) (push (format "%d" (car l)) new-list))
	    ((equal (car l) ""))
	    ((stringp (car l)) (push (car l) new-list))
	    ((listp (car l))
	     (setq new-list (nconc (nreverse (xmh-list-to-string (car l)))
				   new-list)))
	    (t (error "Bad argument %s" (car l))))
      (setq l (cdr l)))
    (nreverse new-list)))


;; (following function from Jim Moy...)
;; Add this function to your xmh-draft-mode-hook and get a nice little header 
;; in your mail messages that look like this: 
;; 
;; 	X-Moon: the moon is waxing gibbous (69% illuminated)
;; 
;; Just for the fun of it.
;; 
;; -jbm
(defun xmh-moon-header ()
  "Updates or creates a header of the form X-Moon in the draft containing
the phase of the moon information returned by Al Silverstein's moon(1)
program called with the -d option."
  (interactive)
  (let ((moon-header "X-Moon: "))
    (goto-char (point-min))
    ;; If there's already one there, delete it before going on.  This means
    ;; the updated one might be in a different place than the old one.
    (if (re-search-forward (concat "^" moon-header) (point-max) t)
	(progn
	  (beginning-of-line)
	  (kill-line 1)))
    ;; Now make the moon entry.
    (goto-char (point-min))
    (re-search-forward "^To:" nil t)
    (cond ((re-search-forward "^--------" nil t)
	   (beginning-of-line)
	   (insert moon-header)
	   (call-process "moon" nil t nil "-d"))
	  (t nil))))


;; copied this from Jim Moy -- I'd like to make a backwards version...
(defun xmh-mail-header-tabs ()
  "Use tab key to go to the next header line, or just tab if we're in the
body of a message."
  (interactive)
  (if (save-excursion
	(re-search-backward "^-*$" (point-min) t 1))
      (insert-tab) ; In body, do normal thing.
    (if (and
	 (re-search-forward "^[-a-zA-Z]+:.*" (point-max) "" 1) ; find header
	 (save-excursion ; presently lacks trailing white space
	 (backward-char)
	 (not (looking-at " ")))) ; then insert trailing blank
	;; else, we're left at the header item (if any), or EOF
	(insert " "))))
	
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; SETUP KEYMAPS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-key ctl-x-map "m" 'xmh-smail)
(define-key xmh-draft-mode-map "\C-C\C-Fb" 'xmh-to-field)
(define-key xmh-draft-mode-map "\C-C\C-Fc" 'xmh-to-field)
(define-key xmh-draft-mode-map "\C-C\C-Ff" 'xmh-to-field)
(define-key xmh-draft-mode-map "\C-C\C-Fs" 'xmh-to-field)
(define-key xmh-draft-mode-map "\C-C\C-Ft" 'xmh-to-field)
(define-key xmh-draft-mode-map "\C-C\C-Q" 'xmh-exit-preserving-draft)
(define-key xmh-draft-mode-map "\C-C\C-K" 'xmh-exit-killing-draft)
(define-key xmh-draft-mode-map "\C-C\C-S" 'xmh-exit-sending-draft)
(define-key xmh-draft-mode-map "\C-C\C-W" 'xmh-check-whom)
(define-key xmh-draft-mode-map "\C-Ci" 'xmh-insert-message-into-draft)
(define-key xmh-draft-mode-map "\C-I" 'xmh-mail-header-tabs)

(provide 'xmh-mode)

