;;;
;;;				NO WARRANTY
;;; 
;;; This software is distributed free of charge and is in the public domain.
;;; Anyone may use, duplicate or modify this program.  Thinking Machines
;;; Corporation does not restrict in any way the use of this software by
;;; anyone.
;;; 
;;; Thinking Machines Corporation provides absolutely no warranty of any kind.
;;; The entire risk as to the quality and performance of this program is with
;;; you.  In no event will Thinking Machines Corporation be liable to you for
;;; damages, including any lost profits, lost monies, or other special,
;;; incidental or consequential damages arising out of the use of this program.
;;;
;;; Thu May  4 1989
;;;
;;; $Header: /tmp_mnt/am/p7/utility/gmacs/f2/RCS/zmacs-compatibility.el,v 1.3 89/02/04 13:06:47 salem Exp Locker: fad $

;;; The following are of interest :
;;;  - SUPER-SPACE 
;;;     (actually bound to C-M-SPACE) rotates through the mark ring
;;;     There are two functions :
;;;       ROTATE-MARK-BACKWARD-COMMAND
;;;       ROTATE-MARK-FORWARD-COMMAND
;;;  - C-M-Y
;;;     repeats the last minibuffer command
;;;     M-Y gets ones further back
;;;  - C-M-I (only if *zmacs-compatibility* is t) 
;;;    same as C-M-Y
;;;  - C-> and C-<
;;;     Marks regions to the end or beginning of the buffer.
;;;     See MARK-REGION-TO-END-OF-BUFFER and 
;;;         MARK-REGION-TO-BEGINNING-OF-BUFFER
;;;  - C-% or c-x %
;;;     Like query replace but replaces each occurance. If given a numeric
;;;   argument, they are only replaced that many times.  Prints the number of
;;;   replacements.
;;;  - M-X Save-File-Buffers
;;;    Same as save-some-buffers (zmacs compatibility)
;;;  - C-M-sh-; and c-X c-; and c-x sh-;
;;;    Comment out region (with a numeric argument deletes the comment char)
;;;    comment-character holds the value of the comment character.
;;;  - c-sh-u
;;;    Undo 
;;;  - M-X Create-Directory
;;;  - M-X Delete-Directory
;;;  - Hardcopy :
;;;     hardcopy-region, hardcopy-buffer, hardcopy-file
;;;  - m-X find-unbalanced-parentheses
;;;  - m-? Describe key
;;;  - c-sh-c, c-sh-e, m-sh-c, m-sh-e
;;;    eval current definition for emacs lisp
;;;  - c-m-c
;;;    Compile emacs lisp buffer
;;;  - buffer switching
;;;    Return in minibuffer completes buffer name
;;;  - c-x 2
;;;    If *zmacs-compatibility* is T, switches to the other buffer in 
;;;  the second window.

;;; Change Log :
;;; 10/6/88 salem : fixed bug in rotate-mark if (mark) returns nil
;;; 10/13/88 salem : fixed c-x 2 implementation to not smash the split 
;;;       window function
;;; 12/27/88 brewster: hardcopy-file handles .dvi files.

(provide 'zmacs)

(defvar *zmacs-compatibility* t
  "*Set this to T to enable zmacs compatibility features that conflict 
with standard gnu emacs bindings.")

;;;***********************************************
;;; More cursor motion
;;;***********************************************

(defun rotate-mark-ring (n)
  (when (> (length mark-ring) 1)
; diag.
;    (dolist (i mark-ring)
;      (when (numberp i) (error "Number in mark ring. (before)"))
;      )
    (cond ((zerop n))
	  ((minusp n)
	   (let ((last (car (last mark-ring))))
	     (setq mark-ring (cons (copy-marker (mark-marker))
				   (butlast mark-ring)))
	     (set-mark last)
	     )
	   (rotate-mark-ring (1+ n))
	   )
	  (t
	   (let ((top (pop mark-ring)))
	     (setf (cdr (last mark-ring)) (list (copy-marker (mark-marker))))
	     (set-mark top)
	     )
	   (rotate-mark-ring (1- n))
	   ))
    (dolist (i mark-ring)
      (when (numberp i) (error "Number in mark ring."))
      )))

(defun rotate-mark-backward-command (n)
  "Rotates back through the mark ring.  With a numeric arg, rotates that many
backwards.  The first time it is called.  It saves the current point as the 
mark."
  (interactive "p")
  (setq this-command 'rotate-mark)
  (when (mark)
    (let ((current-point (point))
	  (current-mark (mark))
	  )
      (rotate-mark-ring (1- n))
      (goto-char (mark))
      ;; Save the current point if necessary
      (when (and (not (eq last-command this-command))
		 (not (eq current-mark current-point))
		 )
	(push-mark current-point t)
	;; skip past it
	(rotate-mark-ring 1)
	)
      ;; Point to next elt.
      (rotate-mark-ring 1)
      ))
  )

(defun rotate-mark-forward-command (n)
  "Rotates forward through the mark ring.  With a numeric arg, rotates that 
many forwards."
  (interactive "p")
  (rotate-mark-backward-command (- n))
  )	       

;;(global-set-key (key-for-chord ?  '(control meta)) 
		'rotate-mark-backward-command)
(global-set-key "\M-\C-@" 'rotate-mark-backward-command)

;;;***********************************************
;;; C-M-Y
;;;***********************************************
(global-set-key "\M-\C-y" 'repeat-complex-command)
(define-key repeat-complex-command-map "\M-y" 'repeat-complex-yank-pop)

;; Add m-c-i so we have a way of doing this in shell mode
;; Only if *zmacs-compatibility* is set  fad 5/16/89
(unless (and (boundp '*zmacs-compatibility*)
	     (null *zmacs-compatibility*))
  (global-set-key "\M-\C-i" 'repeat-complex-command)
  )
(unless (and (boundp '*zmacs-compatibility*)
	     (null *zmacs-compatibility*))
  (define-key repeat-complex-command-map "\M-\C-i" 'repeat-complex-yank-pop)
  )

(defun repeat-complex-yank-pop (n)
  (interactive "p")
  (cond ((eq last-command 'yank) (yank-pop n))
	(t
	 (previous-complex-command n))))


;;;***********************************************
;;; Mark EOB BOB regions
;;;***********************************************

(defun mark-region-to-end-of-buffer ()
  "Marks the region from POINT to the end of the buffer.  Leaves POINT in the 
same place."
  (interactive)
  (let ((save-point (point)))
    (goto-char (point-max))
    (push-mark)
    (goto-char save-point)
    ))

(defun mark-region-to-beginning-of-buffer ()
  "Marks the region from POINT to the beginning of the buffer.  Leaves POINT 
in the same place."
  (interactive)
  (let ((save-point (point)))
    (goto-char (point-min))
    (push-mark)
    (goto-char save-point)
    ))

;;(global-set-key (key-for-chord ?> '(control)) 'mark-region-to-end-of-buffer)
;;(global-set-key (key-for-chord ?< '(control)) 'mark-region-to-beginning-of-buffer) 
;; bug in mapping code
;;(global-set-key (key-for-chord ?. '(control shift)) 'mark-region-to-end-of-buffer)
;;(global-set-key (key-for-chord ?, '(control shift)) 'mark-region-to-beginning-of-buffer) 

;;;***********************************************
;;; Replace
;;;***********************************************

(defvar old-replace-match (symbol-function 'replace-match))
(defvar inside-replace-all nil)
(defvar replace-all-count)
(defvar replace-all-max nil)

(defun replace-match (newtext &optional fixedcase literal)
  (when inside-replace-all
    (when (and replace-all-max (>= replace-all-count replace-all-max))
      (throw 'replace-all-done nil))
    (incf replace-all-count)
    )
  (funcall old-replace-match newtext fixedcase literal)
  )

(defun replace-all-prompt ()
  (let* ((numtimes (and current-prefix-arg (prefix-numeric-value current-prefix-arg)))
	 thing replacement
	 (numtimes-prompt (if numtimes
			      (format "(%d) " numtimes)
			      ""))
	 )
    (setq thing (read-string (concat numtimes-prompt "Replace: ")))
    (setq replacement (read-string (concat numtimes-prompt "Replace " thing " with: ")))
    (list thing replacement numtimes)
    ))


(defun replace-all (from to &optional numtimes)
  "Like query recplace but replaces each occurance of FROM with TO 
with out asking. If given a numeric argument, they are only replaced that
many times.  Prints the number of replacements."
  (interactive (replace-all-prompt))
  (let ((saved-point (point))
	(inside-replace-all t)
	(replace-all-count 0)
	(replace-all-max numtimes)
	)
    (catch 'replace-all-done
      (perform-replace from to nil nil nil))
    (princ (if (= 1 replace-all-count) "1 replacement."
	       (format "%d replacements." replace-all-count)))
    (goto-char saved-point)
    ))

;;(global-set-key (key-for-chord ?% '(control)) 'replace-all)
;; bug in mapping code
;;(global-set-key (key-for-chord ?5 '(control shift)) 'replace-all)
(global-set-key "\C-x%" 'replace-all)

;;;---------------------------------------------------------------------------
;;; Comment out region
;;;---------------------------------------------------------------------------
(defvar comment-character ?\;)

(defun comment-out-region (beg end)
  "Comment out region (with a numeric argument deletes the comment char)
 comment-character holds the value of the comment character."

  (interactive "*r")

  (let ((start-point (point-marker))
	(end-marker (make-marker))
	)
    (goto-char beg)
    (set-marker end-marker end)
    
    (cond (current-prefix-arg
	    (forward-line 0)
	    (while (< (point) (marker-position end-marker))
	      (if (eql comment-character (char-after (point)))
		  (delete-char 1)
		  )
	      (forward-line 1)
	      ))
	  (t
	    (forward-line 0)
	    (while (< (point) (marker-position end-marker))
	      (insert-char comment-character 1)
	      (forward-line 1)
	      )))
    (goto-char (marker-position start-point))
    ))

;;(global-set-key (key-for-chord ?\; '(control meta shift)) 'comment-out-region)
(define-key ctl-x-map ":" 'comment-out-region)
;(define-key ctl-x-map (key-for-chord ?; '(control)) 'comment-out-region)

;;;***************************************************************************
;;; M-X commands
;;;***************************************************************************

(defun save-file-buffers ()
  (interactive)
  (save-some-buffers))

(defun shell-command-fast (string)
  (let ((buf (get-buffer-create "*Shell Command Output*"))
	)
    (save-excursion
      (set-buffer buf)
      (erase-buffer))
    (call-process shell-file-name nil buf nil "-f" "-c" string)
    ))

(defun display-shell-error-if-necessary ()
  "Returns T if there was an error."
  (save-excursion
    (set-buffer (get-buffer "*Shell Command Output*"))
    (cond ((= (point-min) (point-max))
	   ;; no error
	   nil
	   )
	  (t
	   (error (buffer-substring (point-min) (1- (point-max))))
	   t
	   ))))
	   

(defun create-directory (filename)
  "Creates a directory."
  (interactive "FDirectory to create: ")
  (setq filename (expand-file-name filename))
  (cond ((or (null filename) (string-equal filename ""))
	 (error "No default."))
	(t
	 (let* ((len (length filename))
		(last-char (and (plusp len) (aref filename (- len 1))))
		)
	   (when (= ?/ last-char)
	     (setq len (1- len))
	     )
	   (shell-command-fast (concat "/bin/mkdir " (substring filename 0 len)))
	   (unless (display-shell-error-if-necessary)
	     (message "Created directory %s" filename)
	     )
	   ))))

(defun delete-directory (filename)
  "Deletes a directory."
  (interactive "FDirectory to delete: ")
  (setq filename (expand-file-name filename))
  (cond ((or (null filename) (string-equal filename ""))
	 (error "No default."))
	(t
	 (let* ((len (length filename))
		(last-char (and (plusp len) (aref filename (- len 1))))
		)
	   (when (= ?/ last-char)
	     (setq len (1- len))
	     )
	   (shell-command-fast (concat "/bin/rmdir " (substring filename 0 len)))
	   (unless (display-shell-error-if-necessary)
	     (message "Deleted directory %s" filename)
	     )
	   ))))

;;;---------------------------------------------------------------------------
;;; Undo and redo
;;;---------------------------------------------------------------------------

;;(global-set-key (key-for-chord ?u '(control shift)) 'undo)
(if (fboundp 'gmacs-keycode)
    (global-set-key (gmacs-keycode "u" '(control shift)) 'undo)
    )

;;;---------------------------------------------------------------------------
;;; Hardcopy
;;;---------------------------------------------------------------------------

(autoload 'print-region-1 "lpr")

(defun hardcopy-buffer ()
  "Print buffer contents as with Unix command `lpr -p'.  With a numeric arg.,
suppress page headers (via the -B switch).  `lpr-switches' is a list 
of extra switches (strings) to pass to lpr."   
  (interactive)
  (print-region-1 (point-min) (point-max)
		  (append (if current-prefix-arg '("-B"))
			  (cons "-p" lpr-switches))))

(defun hardcopy-region (start end)
  "Print region contents as with Unix command `lpr -p'.  With a numeric arg.,
suppress page headers (via the -B switch).  `lpr-switches' is a list 
of extra switches (strings) to pass to lpr."   
  (interactive "r")
  (print-region-1 start end
		  (append (if current-prefix-arg '("-B"))
			  (cons "-p" lpr-switches))))

(defun hardcopy-file (file)
  "Print file contents as with Unix command `lpr -p'. 
   If the file is a .dvi file, then it uses `lpr -d'.
   With a numeric arg.,  suppress page headers (via the -B switch).
   `lpr-switches' is a list 
   of extra switches (strings) to pass to lpr."   
  (interactive "FHardcopy file: ")
  (setq file (expand-file-name file))
  (let ((file-type (file-type file)))
    (cond ((null file-type)
	   (print-files file (append (if current-prefix-arg '("-B"))
		 (cons "-p" lpr-switches))))
	  ((equal "dvi" file-type)
	   (print-files file (cons "-d" lpr-switches)))
	  ;;this should handle tex files well.  another day, another hack.
	  (t (print-files file (append (if current-prefix-arg '("-B"))
				       (cons "-p" lpr-switches)))
	     ))))

(defun file-type (filename)
  (let ((answer (zmacs-string-reverse-search  "\\." filename)))
    (if answer
	(substring filename (1+ answer))
	nil)))

(defun zmacs-string-reverse-search (character string)
  "this searches from the back of a string for an occurance of key.
   this should be written in C and not cons.  sigh.
  "
  (let ((answer (string-match character (reverse-string string))))
    (if answer
	(- (length string) answer 1)
	nil)))

(defun reverse-string (string)
  (let* ((length (length string))
	 (result (make-string length 0)))
    (dotimes (i (length string))
      (setf (aref result i) (aref string (- length i 1))))
    result))

(defun regexp-quote-for-directory-list (filename)
  (let (pos)
    (setq filename (regexp-quote filename))
    (while (setq pos (string-match "\\\\\\*" filename pos))
      (aset filename pos ?.)
      )
    (concat "^" filename "$"))
  )

;;; (regexp-quote-for-directory-list "abcde") --> ^abcde$
;;; (regexp-quote-for-directory-list "ab.cde") --> ^ab\\.cde$
;;; (regexp-quote-for-directory-list "ab*cde") --> ^ab.*cde$

(defun print-files (file switches)
  "File may have wildcards."
  (let ((printedp nil))
    (cond ((string-match "\\*" file)
	   (let ((list
		   (directory-files (file-name-directory file)
				    t
				    (regexp-quote-for-directory-list
				      (file-name-nondirectory file))))
		 )
	     (while list
	       (print-file-1 (car list) switches)
	       (setq printedp t)
	       (setq list (cdr list))
	       )
	     ))
	  ((file-exists-p file)
	   (print-file-1 file switches)
	   (setq printedp t)
	   ))
    (unless printedp	  
      (error "File %s doesn't exist." file))
    ))

(defun print-file-1 (file switches)
  (save-excursion
   (message "Spooling %s ..." file)
   (if (/= tab-width 8)
       (let ((oldbuf (current-buffer)))
	(set-buffer (get-buffer-create " *spool temp*"))
	(widen) (erase-buffer)
	(insert-buffer-substring oldbuf)
	(call-process "expand" t t nil
		      (format "-%d" tab-width) file)))
   (apply 'call-process
	  (append (list (if (eq system-type 'usg-unix-v)
				     "lp" "lpr")
			nil nil nil
			"-J" file "-T" file)
		  switches
		  (list file)
		  ))
   (message "Spooling %s ...done" file)))

;;;---------------------------------------------------------------------------
;;; Find unbalanced
;;;---------------------------------------------------------------------------

(defun find-unbalanced-parentheses ()
  "Check the buffer for unbalanced parentheses.  Stops at any that are 
unbalanced."
  (interactive)
  (let ((start-point (point))
	)
    (goto-char (point-min))
    (condition-case e
	 (while (/= (point) (point-max))
	   (forward-sexp)
	   )
      (error
	;; If this is an extra left paren error, we have to scan backwards to 
	;; find the exact left paren in error
	(cond ((and (eq (car e) 'error)
		    (string-equal (cadr e) "Unbalanced parentheses")
		    )
	       ;; left paren error
	       (goto-char (point-max))
	       (while (/= (point) (point-min))
		 (condition-case e (backward-sexp)
		    (error
		      (error "Probably an extra left parenthesis here.")))
		 ))
	      (t
	       (error "Probably an extra right parenthesis here.")
	       ))
	))
    (goto-char start-point)
    (message "All parentheses appear balanced.")
    ))
	       

;;;---------------------------------------------------------------------------
;;; Split window
;;;---------------------------------------------------------------------------

(unless (and (boundp '*zmacs-compatibility*)
	     (null *zmacs-compatibility*))
  (global-set-key "\C-x2" 'split-window-vertically-other-buffer)
  )

(defun split-window-vertically-other-buffer (arg)
  "Split current window into two windows, one above the other.
This window becomes the uppermost of the two, and gets
ARG lines.  No arg means split equally.
This selects the (other-buffer) in the other window."
  (interactive "P")
  (split-window-vertically arg)
  (other-window 1)
  (switch-to-buffer (other-buffer))
  )

;;;------------------------------------------------
;;; Describe Key
;;;------------------------------------------------

(global-set-key "\M-?" 'describe-key)

;;;***********************************************
;;; C-sh-c C-sh-e c-m-c for emacs lisp
;;;***********************************************

(defun tmc-eval-current-buffer ()
  "Evaluates all the lisp expressions in the current buffer."
  (interactive)
  (message "Evaluating GMACS buffer . . .")
  (eval-current-buffer)
  (message "Evaluating GMACS buffer . . . Done")
  )

(defun tmc-eval-defun (&optional arg)
  "Evaluates the current definition.  With an argument, it evaluates the 
region."
  (interactive "P")
  (cond (arg
	 (message "Evaluating GMACS region . . .")
	 (eval-region (min (point) (mark)) (max (point) (mark)) nil)
	 (message "Evaluating GMACS region . . . Done")
	 )
	(t
	 (eval-defun nil))
	))

;;; Standard ASCII
(define-key emacs-lisp-mode-map "\M-C" 'tmc-eval-defun)
(define-key emacs-lisp-mode-map "\M-E" 'eval-last-sexp)
(define-key emacs-lisp-mode-map "\M-\C-c" 'tmc-eval-current-buffer)

;;; X10
;;(define-key emacs-lisp-mode-map (key-for-chord ?c '(control shift))
  'tmc-eval-defun)

;;(define-key emacs-lisp-mode-map (key-for-chord ?e '(control shift))
  'eval-last-sexp)

;;; Ann Arbor
(when (fboundp 'gmacs-keycode)
  (define-key emacs-lisp-mode-map (gmacs-keycode "c" '(control shift))
    'tmc-eval-defun)

  (define-key emacs-lisp-mode-map (gmacs-keycode "c" '(control meta))
    'tmc-eval-current-buffer)
  )


	 

;;;-----------------------------------------------
;;;Return in minibuffer completes buffer name
;;;-----------------------------------------------

(global-set-key "\C-xb" 'completing-switch-to-buffer)

(defun completing-switch-to-buffer ()
  (interactive)
  (unwind-protect
       (progn
	 (define-key minibuffer-local-completion-map "\r" 'my-minibuffer-complete-and-exit)
	 (call-interactively 'switch-to-buffer))
    (define-key minibuffer-local-completion-map "\r" 'exit-minibuffer)
    ))

(defun my-minibuffer-complete-and-exit  ()
  (interactive)
  (let ((minibuffer-completion-confirm nil))
    (minibuffer-complete-and-exit)
    ))

